/*
 Copyright (C) 2011 Christian Dywan <christian@twotoasts.de>

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 See the file COPYING for the full license text.
*/

namespace Postler {

    [DBus (name = "org.elementary.Postler")]
    interface PostlerClient : Object {
        public async abstract int64[] unread_messages (string uri) throws IOError;
        public async abstract GLib.HashTable<string,Variant> get_message (string id,
            char flag) throws IOError;
        public async abstract string[] get_thread (string thread) throws IOError;
        public signal void got_message (GLib.HashTable<string,Variant> message, int64 position);
        public async abstract string[] get_messages (string uri) throws IOError;
        public signal void progress (string account, string text, double fraction);
        public abstract void receive (string account) throws IOError;
        public signal void received (string account, string error_message);
        public abstract bool fetch (string account) throws IOError;
        public abstract void send (string account, string filename) throws IOError;
        public signal void sent (string account, string filename, string error_message);
        public abstract void quit () throws IOError;
    }

    public class Client : Object {
        static PostlerClient? client = null;

        public Client () {
            if (client != null)
                return;
            try {
                client = Bus.get_proxy_sync (BusType.SESSION,
                                             "org.elementary.Postler",
                                             "/org/elementary/postler");
                client.got_message.connect ((message_data, position) => {
                    var message = new Message.from_id (message_data.lookup ("id").get_string ());
                    message.from_hash_table (message_data);
                    got_message (message, position);
                });
                client.progress.connect ((account, text, fraction) => {
                    progress (account, text != "" ? text : null, fraction);
                });
                client.received.connect ((account, error_message) => {
                    received (account, error_message != "" ? error_message : null);
                });
                client.sent.connect ((account, filename, error_message) => {
                    sent (account, filename, error_message != "" ? error_message : null);
                });
                /* Ensure Postler is running, ignore errors */
                Process.spawn_async (null,
                    { Postler.App.argv0, "-m", "service" }, null,
                    SpawnFlags.SEARCH_PATH,
                    null, null);
            } catch (GLib.Error error) { }
        }

        public async int64[] unread_messages (string uri) throws GLib.Error {
            int64[] new_messages = { 0, 0 };
            return_val_if_fail (uri != null, new_messages);
            return yield client.unread_messages (uri);
        }

        public async void get_message (Message message, char flag,
            Gtk.TreeRowReference row) {

            try {
                message.from_hash_table (yield client.get_message (message.id, flag));
            } catch (GLib.Error error) {
                message.from_error (error.message.replace (
    "org.gtk.GDBus.UnmappedGError.Quark._g_2dfile_2derror_2dquark.Code24", "GLib.FileError"));
            }
            /* The path is null if the row was removed during the async call */
            Gtk.TreePath? path = row.get_path ();
            if (path == null)
                return;
            Gtk.TreeIter iter;
            if (row.get_model ().get_iter (out iter, path))
                row.get_model ().row_changed (path, iter);
        }

        public async GLib.List<Message> get_thread (Message parent) throws GLib.Error {
            var messages = new GLib.List<Message> ();
            try {
                string[] ids = yield client.get_thread (parent.id);
                foreach (string message_id in ids) {
                    var child = new Message.from_id (message_id);
                    child.from_hash_table (yield client.get_message (child.id, MessageFlags.NONE));
                    messages.append (child);
                }
            }
            catch (GLib.Error error) {
                var message = new Message.from_id ("error:");
                message.from_error (error.message.replace (
    "org.gtk.GDBus.UnmappedGError.Quark._g_2dfile_2derror_2dquark.Code24", "GLib.FileError"));
                messages.append (message);
            }
            return messages;
        }

        public signal void got_message (Message message, int64 position);

        public async GLib.List<Message> get_messages (string uri) throws GLib.Error {
            var messages = new GLib.List<Message> ();
            try {
                string[] ids = yield client.get_messages (uri);
                foreach (string message_id in ids)
                    messages.append (new Message.from_id (message_id));
            }
            catch (GLib.Error error) {
                var message = new Message.from_id ("error:");
                message.from_error (error.message.replace (
    "org.gtk.GDBus.UnmappedGError.Quark._g_2dfile_2derror_2dquark.Code24", "GLib.FileError"));
                messages.append (message);
            }
            return messages;
        }

        public signal void progress (string account, string? text, double fraction);

        public void receive (string account="") {
            try {
                client.receive (account);
            } catch (GLib.Error error) {
                received (account, error.message);
            }
        }

        public signal void received (string account, string? error_message);

        public bool fetch (string account="") {
            try {
                return client.fetch (account);
            } catch (GLib.Error error) {
                return false;
            }
        }

        public void send (string account, string filename) {
            try {
                client.send (account, filename);
            } catch (GLib.Error error) {
                sent (account, filename, error.message);
            }
        }

        public signal void sent (string account, string filename, string? error_message);

        public void quit () {
            try {
                client.quit ();
            } catch (GLib.Error error) {
            }
        }
    }
}

