/*
 Copyright (C) 2011 Christian Dywan <christian@twotoasts.de>

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 See the file COPYING for the full license text.
*/

struct Postler.EmoticonMapping {
    public string token;
    public string icon_name;
}

public class Postler.Content : WebKit.WebView {
    public Message message { get; private set; }
    string? save_folder = null;

    public string? last_location { get; set; }
    public string? message_id { get; set; }
    public string? sender { get; set; }
    public string? subject { get; set; }
    public string? reply_to { get; set; }
    public string? list_post { get; set; }
    public string? list_unsubscribe { get; set; }

    public unowned List<MessagePart> message_parts { public get; set; }
    public MessagePart? current_part { get; private set; }
    public int current_part_index { get; private set; }

    Gtk.Settings gtk_settings;
    GLib.Settings glib_settings;

    const string fallback_style_sheet = """
        body {
            background-color: Window !important; color: WindowText !important;
        }
        .header {
            background-color: Window !important; color: WindowText !important;
        }
        .reply, .status {
            border-radius: 15px;
            border: 1px solid ButtonText;
            background-color: ButtonFace; color: ButtonText;
            padding: 1em; margin: 1em;
        }
        .time {
            float: right;
        }
        .sender {
            margin-bottom: 0.5em;
        }
        .avatar {
            display: none;
        }
        .avatar[src^=file], .avatar[src^=http] {
            display: inline;
            width: 48px; height: 48px;
            float: right;
        }
        .button, .button[href] {
            padding: 0.4em 1em !important;
            -webkit-appearance: push-button !important;
            color: ButtonText !important;
            text-decoration: none !important;
        }
        /* Addresses not underlined, but underlined when hovering */
        a {
            color: Link !important;
        }
        a[href] {
            text-decoration: underline !important;
        }
        a[href]:hover {
            text-decoration: none !important;
        }
        """;

    const string style_sheet = """
        .headers {
            display: block;
            font-size: 100% !important;
            margin: 0em; padding: 0.5em;
        }
        * {
            padding: 0em; margin: 0em;
        }
        body {
            padding: 1em; margin: 0 !important;
        }
        .body {
            font-size: 100% !important;
        }
        .no_signature .signature {
            display: none;
        }
        /* Quotations */
        blockquote {
            margin: 1em;
            padding: 0.5em;
            white-space: pre-wrap;
            color: GrayText;
        }
        /* Plain text */
        .plain_text {
            white-space: pre-wrap;
            font-family: monospace;
        }
        /* Signature */
        .signature {
            color: GrayText;
        }
        .signature:before {
            content: "——";
        }
        .actions {
            float: right;
        }
        /* Dynamically sized and visible statusbar */
        a[href^=http]:hover:after {
            content: attr(href);
            position: fixed; left: 1pt; bottom: 1pt;
            padding: 0 1pt !important; font-weight: normal !important;
            max-width: 95%; overflow: hidden;
            white-space: nowrap; text-overflow: ellipsis;
            font-size: 100% !important; text-shadow: 0 0 5pt ButtonHighlight;
            background-color: ButtonFace !important; color: ButtonText !important;
            outline: ButtonFace solid thick; z-index: 9999;
        }
        /* Contain images to the size of the view */
        img {
            max-height:100%; max-width:100%;
        }
        pre { white-space: pre-wrap; }
        """;

    void font_name_changed (GLib.Object gtk_settings, GLib.ParamSpec pspec) {
        string gtk_font_name;
        gtk_settings.get ("gtk-font-name", out gtk_font_name);
        var font = Pango.FontDescription.from_string (gtk_font_name);
        settings.set ("default-font-family", font.get_family (),
                      "cursive-font-family", font.get_family (),
                      "fantasy-font-family", font.get_family (),
                      "default-font-size", font.get_size () / Pango.SCALE);
    }

    void monospace_font_changed (GLib.Settings glib_settings, string key) {
        string monospace_font = glib_settings.get_string (key);
        var font = Pango.FontDescription.from_string (monospace_font);
        settings.set ("monospace-font-family", font.get_family (),
                      "default-monospace-font-size", font.get_size () / Pango.SCALE);
    }

    public Content () {
        settings.set ("enable-scripts", false, "enable-plugins", false);
        if (settings.get_class ().find_property ("enable-private-browsing") != null) {
            settings.set ("enable-private-browsing", true);
        }
        navigation_policy_decision_requested.connect (navigation_decision);
        create_web_view.connect (new_window_created);
        populate_popup.connect (populate_menu);

        gtk_settings = Gtk.Settings.get_for_screen (get_screen ());
        gtk_settings.notify["gtk-font-name"].connect (font_name_changed);
        font_name_changed (gtk_settings,
            new GLib.ParamSpecString ("gtk-font-name", "", "", "", 0));
        foreach (string schema in GLib.Settings.list_schemas ()) {
            if (schema != "org.gnome.desktop.interface")
                continue;
            glib_settings = new GLib.Settings (schema);
            glib_settings.changed["monospace-font-name"].connect (monospace_font_changed);
            monospace_font_changed (glib_settings, "monospace-font-name");
        }
    }

    internal static string format_x_mailer (string x_mailer) {
        string mailer = x_mailer.chomp ();
        if (mailer.has_prefix ("Mozilla/") && ") " in mailer)
            return mailer.rchr (-1, ' ').chug ().replace ("/", " ");

        int bracket = 0;
        while (mailer[bracket] != '(' && mailer[bracket] != '\0')
            bracket++;
        if (mailer[bracket] == '\0')
            return mailer;
        return mailer.substring (0, bracket).chomp ().replace ("/", " ");
    }

    static string html_escape (string unescaped) {
        return unescaped.replace ("\"", "&quot;")
            .replace ("<", "&lt;").replace (">", "&gt;");
    }

    string linkify_icon (string icon_name, string label, string uri) {
        string markup = label;
        string? real_icon_name = null;
        var screen = Gdk.Screen.get_default ();
        if (screen != null) {
            var icon_theme = Gtk.IconTheme.get_for_screen (screen);
            if (icon_theme.has_icon (icon_name + "-symbolic"))
                real_icon_name = icon_name + "-symbolic";
            else if (icon_theme.has_icon (icon_name))
                real_icon_name = icon_name;
            markup = "<img src=\"%s\" title=\"%s\">".printf (
                icon_name_to_data_uri (real_icon_name, Gtk.IconSize.BUTTON), label);
        }
        return "<a href=\"%s\" class=\"button\">%s</a>".printf (uri, markup);
    }

    internal static string linkify_address (string mailto, bool inherit=false) {
        var linkified = new StringBuilder ();

        string[] parts = mailto.split ("?", 2);
        string addresses = parts[0];
        string arguments = parts[1] != null ? ("?" + parts[1]) : "";

        int count = 0;
        foreach (string address in Contact.quoted_split (addresses)) {
            if (address == "")
                continue;
            if (!address.contains ("@")) {
                linkified.append (html_escape (address));
                continue;
            }
            string[] parsed = Postler.Contact.parse (address);
            string name = html_escape (parsed[0]);
            string quoted = html_escape (address);
            count++;
            if (count > 12)
                continue;
            if (linkified.len > 0)
               linkified.append (", ");
            linkified.append_printf ("<a href=\"mailto:%s%s\"", quoted, arguments);
            if (inherit)
                 linkified.append (" style=\"color:inherit\"");
            linkified.append ("\" title=\"" + quoted + "\">" + name + "</a>");
        }
        /* FIXME: provide a way to see those others */
        if (count > 12)
            linkified.append_printf (_(" and %d others"), count - 12);
        return linkified.str;
    }

    void populate_menu (Gtk.Menu menu) {
        if (editable || can_copy_clipboard ())
            return;

        menu.hide ();

        var event = Gtk.get_current_event ();
        var result = get_hit_test_result ((Gdk.EventButton?)event);
        Gtk.MenuItem menuitem;

        if (result.link_uri != null) {
            menuitem = new Gtk.ImageMenuItem.with_mnemonic (_("Copy _Address"));
            var icon = new Gtk.Image.from_stock (Gtk.STOCK_COPY, Gtk.IconSize.MENU);
            (menuitem as Gtk.ImageMenuItem).set_image (icon);
            menuitem.activate.connect ((menuitem) => {
                var clipboard = get_clipboard (Gdk.SELECTION_CLIPBOARD);
                string address = result.link_uri;
                if ("@" in address)
                    clipboard.set_text (address.split ("?")[0], -1);
                else
                    clipboard.set_text (address, -1);
            });
            menuitem.show ();
            menu.append (menuitem);
            return;
        }

        menuitem = new Gtk.ImageMenuItem.from_stock (Gtk.STOCK_COPY, null);
        menuitem.activate.connect ((menuitem) => {
            copy_clipboard (); });
        menuitem.show ();
        menuitem.sensitive = can_copy_clipboard ();
        menu.append (menuitem);
        menuitem = new Gtk.MenuItem.with_mnemonic (_("Copy _Filename"));
        menuitem.activate.connect ((menuitem) => {
            var clipboard = get_clipboard (Gdk.SELECTION_CLIPBOARD);
            clipboard.set_text (last_location, -1);
        });
        menuitem.show ();
        menu.append (menuitem);

        menuitem = new Gtk.SeparatorMenuItem ();
        menuitem.show ();
        menu.append (menuitem);
        menuitem = new Gtk.ImageMenuItem.from_stock (Gtk.STOCK_ZOOM_IN, null);
        menuitem.label = _("_Enlarge Text");
        menuitem.activate.connect ((menuitem) => {
            zoom_in ();
        });
        menuitem.show ();
        menu.append (menuitem);
        menuitem = new Gtk.ImageMenuItem.from_stock (Gtk.STOCK_ZOOM_OUT, null);
        menuitem.label = _("Sh_rink Text");
        menuitem.activate.connect ((menuitem) => {
            zoom_out ();
        });
        menuitem.show ();
        menu.append (menuitem);

        if (get_view_source_mode ())
            return;

        menuitem = new Gtk.SeparatorMenuItem ();
        menuitem.show ();
        menu.append (menuitem);
        menuitem = new Gtk.MenuItem.with_mnemonic (_("View _Source"));
        menuitem.activate.connect ((menuitem) => {
            Postler.App.spawn_module ("source", last_location); });
        menuitem.show ();
        menuitem.sensitive = last_location != null;
        menu.append (menuitem);
        menuitem = new Gtk.ImageMenuItem.from_stock (Gtk.STOCK_PRINT, null);
        menuitem.activate.connect ((menuitem) => {
            get_main_frame ().print (); });
        menuitem.show ();
        menuitem.sensitive = last_location != null;
        menu.append (menuitem);
    }

    public virtual signal void clear () {
        last_location = null;
        reply_to = null;
        message_parts = null;
        notify_property ("message-parts");
        load_string ("<style text=\"text/css\">%s</style><body></body>"
                     .printf (themed_style_sheet ()),
                     "text/html", "UTF-8", "about:blank");
    }

    const EmoticonMapping[] emoticons = {
        { ":-D", "face-smile-big" },
        { ":-(", "face-sad" },
        { ":'-(", "face-crying" },
        { "O:-)", "face-angel" },
        { ":-|", "face-plain" },
        { ":-*", "face-kiss" },
        { ":-)", "face-smile" },
        { ";-)", "face-wink" },
        { ":-O", "face-surprise" },
        { "8-)", "face-glasses" },
        { ":-P", "face-monkey" },
        { ":)", "face-smile" },
        { "xD", "face-smile-big" },
        { ";)", "face-wink" },
        { "^_^", "face-smile" },
        { "~_^", "face-wink" },
        { "^_~", "face-wink" }
    };

    static string? icon_name_to_data_uri (string icon_name,
        Gtk.IconSize icon_size=Gtk.IconSize.MENU) {

        Gdk.Pixbuf? pixbuf = null;
        var toplevels = Gtk.Window.list_toplevels ();
        if (toplevels.nth_data (0) != null)
            pixbuf = toplevels.nth_data (0).render_icon (icon_name, icon_size, null);
        if (pixbuf == null) {
            try {
                pixbuf = Gtk.IconTheme.get_for_screen (
                    Gdk.Screen.get_default ()).load_icon (icon_name,
                        icon_size, Gtk.IconLookupFlags.FORCE_SIZE);
            } catch (GLib.Error error) { }
        }
        if (pixbuf == null) {
            return null;
        }
        return pixbuf_to_data_uri (pixbuf);
    }

    static string? pixbuf_to_data_uri (Gdk.Pixbuf pixbuf) {
        string buffer;
        size_t buffer_size;
        try {
            if (!Gdk.pixbuf_save_to_buffer (pixbuf, out buffer,
                out buffer_size, "png", null, null)) {
                return null;
            }
        }
        catch (GLib.Error error) {
            return null;
        }

        string encoded = GLib.base64_encode (buffer, buffer_size);
        return "data:image/png;base64," + encoded;
    }

    static bool evaluate_emoticon (GLib.MatchInfo     match_info,
                                   GLib.StringBuilder result,
                                   EmoticonMapping    emoticon) {
        string? data_uri = icon_name_to_data_uri (emoticon.icon_name);
        if (data_uri == null) {
            result.append (" " + emoticon.token);
            return true;
        }

        result.append (" <img src=\"" + data_uri + "\" " +
                       "title=\"" + emoticon.token + "\"/>");
        return false;
    }

    const string[] link_formats = {
        "(https?:\\/\\/[^\\s\\<]+)",
        "([a-zA-Z0-9.\\-]+@[a-zA-Z0-9.\\-]+[a-zA-Z0-9.]+)"
    };

    static string parse_encoded (string quoted, out string charset) {
        return Postler.Messages.parse_encoded (quoted, out charset);
    }

    string format_header (string header, string? data) {
        if (data != null)
            return "<br><b>%s</b> %s<br>".printf (header, data);
        return "";
    }

    internal static string format_date (DateTime the_time) {
        var now = new DateTime.now_local ();
        if (the_time.get_day_of_month () == now.get_day_of_month ())
            return _("Today") + the_time.format (" %X");
        if (the_time.get_day_of_year () == now.get_day_of_year () - 1)
            return _("Yesterday") + the_time.format (" %X");
        if (the_time.get_year () == now.get_year ()) {
            /* i18n: strftime format for full month name and day number */
            return the_time.format (_("%B %e") + (" %X"));
        }
        /* i18n: strftime format for full month name, day number and year */
        return the_time.format (_("%B %e, %Y") + (" %X"));
    }

    internal static DateTime date_from_string (string date) {
        var parsed = new Soup.Date.from_string (date);
        TimeVal time = new TimeVal ();
        int offset = 0;
        if (parsed != null) {
            parsed.to_timeval (time);
            offset = parsed.offset;
        }
        return new DateTime.from_timeval_local (time).add_minutes (offset);
    }

    public void prepare_reply (string? location, AccountInfo account_info,
        string recipient, bool quote, int part, string? excerpt, bool is_bug_tracker) {

        var body = new StringBuilder ();

        if (location != null && !is_bug_tracker) {
            ulong now = new Soup.Date.from_now (0).to_time_t ();
            string date = GLib.Time.local ((time_t)now).format ("%x %X");
            string sender = Postler.Contact.name_from_string (recipient);

            body.append_printf (_("On %s, %s wrote:"), date, sender);
            body.append_c ('\n');
            var contents = File.new_for_path (location);
            try {
                last_location = location;
                message = new Message.from_file (contents);
                message_id = message.id;
                subject = message.subject;
                message.parse_body ();
                message_parts = message.parts.copy ();
                notify_property ("message-parts");

                bool last_line_empty = true;
                StringBuilder b_temp = new StringBuilder ();
                if (excerpt != null)
                    b_temp.append (excerpt);
                else if (mime_type_is_text ((message_parts.nth_data (part)).mime_type)) {
                    b_temp.append ((message_parts.nth_data (part)).get_plain_text ());
                } else {
                    foreach (var bpart in message_parts) {
                        if (mime_type_is_text (bpart.mime_type)) {
                            if (!bpart.is_empty ()) {
                                b_temp.append_c ('\n');
                                b_temp.append (bpart.get_plain_text ());
                                break;
                            }
                        }
                    }
                }
                string [] lines = b_temp.str.split ("\n");
                foreach (var line in lines) {
                    /* Skip effectively empty lines */
                    int position = 0;
                    do {
                        if (line[position] != '>' && line[position] != ' ')
                            break;
                        position++;
                    } while (true);
                    string effective = line.substring (position);
                    if (effective == "") {
                        if (last_line_empty)
                            continue;
                        last_line_empty = true;
                        body.append ("> \n");
                        continue;
                    }
                    /* Skip non-helpful decorative lines */
                    if (effective == "-----BEGIN PGP SIGNED MESSAGE-----"
                     || effective.has_prefix ("Hash: SHA")
                     || effective == "-----END PGP SIGNATURE-----"
                     || effective == "-----BEGIN PGP SIGNATURE-----")
                        continue;
                    last_line_empty = false;
                    string[] words_in_line = line.split (" ");
                    long len = 0;
                    body.append ((quote ? "> " : ""));
                    for (int i = 0; i < words_in_line.length; i++)
                    {
                        len += words_in_line[i].length + 1;
                        if (len < (quote ? 70 : 72)) {
                            body.append (words_in_line[i]);
                            body.append_c (' ');
                        } else {
                            body.append_c ('\n');
                            body.append ((quote ? "> " : ""));
                            body.append (words_in_line[i]);
                            body.append_c (' ');
                            len = 0;
                        }
                    }
                    body.append ("\n");
                }
            } catch (GLib.Error error) {
                display_error (_("Failed to quote message \"%s\": %s").printf (
                    contents.get_path (), error.message));
            }
        }

        /* If the original message was top-post, we also top-post.
           Presumably more quotes than none at the bottom indicates that. */
        bool top_post = false;
        ssize_t position = body.len;
        uint line_breaks = 0;
        uint quote_lines = 0;
        do {
            if (body.str[position] == ' ' && body.str[position - 1] == '>') {
                quote_lines++;
                if (quote_lines > 5) {
                    top_post = true;
                    break;
                }
                position--;
            } else if (body.str[position] == '\n') {
                line_breaks++;
                if (line_breaks > 5)
                    break;
            }
            position--;
        } while (position > 0 && position < body.len);

        if (!is_bug_tracker) {
            unowned string? signature = account_info.signature;
            if (signature != null && top_post)
                body.prepend ("\n\n" + signature + "\n");
            else if (signature != null)
                body.append ("\n\n" + signature);
        }
        load_string ("""
            <style text=\"text/css">
            %s * { white-space: pre-wrap !important; }</style>
            <body>%s</body>
            """.
            printf (themed_style_sheet (), body.str),
            "text/html", "UTF-8", "about:blank");
    }

    bool mime_type_is_text (string mime_type) {
        return g_content_type_is_a (mime_type, "text/plain");
    }

    public async void display_source (Message message) {
        last_location = message.get_path ();
        subject = _("Source Code: %s").printf (last_location);

        string line;
        string content_type = null;
        string mime_type = "text/plain";
        string charset = null;
        string[] parts;

        var contents = File.new_for_path (last_location);
        GLib.StringBuilder body = new GLib.StringBuilder ();

        try {
            var stream = new DataInputStream (contents.read (null));
            while ((line = stream.read_line (null, null)) != null) {
                parts = line.split (":", 2);
                if (parts != null && parts[0] != null) {
                    string field = ascii_strdown (parts[0]);
                    if (field == "content-type")
                        content_type = parts[1].strip ();
                    else if (field == "subject")
                        subject = _("Source Code: %s").printf (parse_encoded (parts[1], out charset));
                }
                body.append (line + "\n");
            }
        } catch (GLib.Error error) {
            display_error (_("Failed to view source: %s").printf (error.message));
            return;
        }

        string? boundary = null;
        string? fname = null;
        Postler.Message.parse_content_type (content_type, ref charset, ref boundary,
                            ref mime_type, ref fname);
        load_string ("""
            <style text="text/css">
            %s * { white-space: pre-wrap !important; }</style>
            </style><body>%s</body>
            """.
            printf (themed_style_sheet (), body.str),
            "text/html", charset, "about:blank");
    }

    public string? list_or_all () {
        if (list_post == null)
            return message.recipients + "," + message.reply_to ?? message.sender;

        if (!list_post.has_prefix ("NO "))
            return list_post;

        var dialog = new Gtk.MessageDialog (get_toplevel () as Gtk.Window, 0,
            Gtk.MessageType.WARNING, Gtk.ButtonsType.NONE,
            _("This message was sent to a mailing list."));
        dialog.format_secondary_text (_("Replying is not allowed."));
        dialog.run ();
        dialog.destroy ();
        return null;
    }

    public string? choose_from (Message the_message) {
        /* See if recipient is among the accounts, otherwise pick a fallback */
        var accounts = new Accounts ();
        string? recipient = the_message.get_field ("to");
        if (recipient != null) {
            string from = Postler.Contact.address_from_string (recipient);
            foreach (unowned AccountInfo info in accounts.get_infos ()) {
                if (info.address == null)
                    continue;
                foreach (string address in info.address.split (",")) {
                    if (Contact.equal (from, address))
                        return from;
                }
            }
        }

        var account_info = the_message.get_account (accounts);
        if (account_info != null)
            return account_info.address.split (",")[0];

        return null;
    }

    public string reply_uri (Message the_message, string? prefix=null, string? recipients=null) {
        string chosen_from = choose_from (the_message);
        string to = recipients ?? (the_message.reply_to ?? the_message.sender);
        StringBuilder mailto = new StringBuilder (to);
        char delimiter = '?';
        if (prefix != null) {
            mailto.append_printf ("%csubject=%s%s",
                delimiter,
                GLib.Uri.escape_string (prefix, "", true),
                GLib.Uri.escape_string (the_message.subject, "", true));
            delimiter = '&';
        }
        if (chosen_from != null) {
            mailto.append_printf ("%cfrom=%s", delimiter, chosen_from);
            delimiter = '&';
        }
        mailto.append_printf ("%cin-reply-to=%s", delimiter, the_message.id);
        if (can_copy_clipboard ()) {
            var clipboard = get_clipboard (Gdk.SELECTION_PRIMARY);
            string? selected = GLib.Uri.escape_string (clipboard.wait_for_text (), "", true);
            mailto.append_printf ("&body=%s", selected);
        }
        return mailto.str;
    }

    public void compose_message (string? prefix=null, string? recipients=null) {
        string mailto = reply_uri (message, prefix, recipients);
        Postler.App.spawn_module ("compose", mailto);
    }

    public async bool display (Message the_message, AccountInfo? account_info=null) {
        try {
            GLib.List<Message> thread;
            if (the_message.stream != null) {
                thread = new GLib.List<Message> ();
                thread.append (the_message);
            }
            else {
                var client = new Postler.Client ();
                thread = yield client.get_thread (the_message);
            }

            /* Use Adium style if existant, otherwise fallback to internal */
            string theme_name = Environment.get_variable ("POSTLER_ADIUM_THEME") ?? "ubuntu";
            string template_dir = "%s/adium/message-styles/%s.AdiumMessageStyle/Contents/Resources/";
            string? template_path = null;
            foreach (string data_dir in Environment.get_system_data_dirs ()) {
                string template_folder = template_dir.printf (data_dir, theme_name);
                if (FileUtils.test (template_folder, FileTest.EXISTS)) {
                    template_path = template_folder;
                    break;
                }
            }
            if (template_path == null)
                template_path = template_dir.printf (
                    Environment.get_user_data_dir (), theme_name);
            string content_template, outgoing_template, status_template, content_stylesheet;
            string? next_content_template = null;
            try {
                FileUtils.get_contents (template_path + "Incoming/Content.html",
                                        out content_template, null);
                try {
                    FileUtils.get_contents (template_path + "Incoming/NextContent.html",
                                            out next_content_template, null);
                }
                catch (GLib.Error next_content_error) { }
                try {
                    FileUtils.get_contents (template_path + "Outgoing/Content.html",
                                            out outgoing_template, null);
                }
                catch (GLib.Error outgoing_error) {
                    outgoing_template = content_template;
                }
                FileUtils.get_contents (template_path + "Status.html",
                                        out status_template, null);
                FileUtils.get_contents (template_path + "main.css",
                                        out content_stylesheet, null);
            }
            catch (GLib.Error template_error) {
                content_template = outgoing_template = """
                    <div class="reply">
                    <div class="sender">%sender%<div class="time">%time{%X}%</div></div>
                    <img src="%userIconPath%" class="avatar">%message%
                    </div>
                    </div>
                    """;
                status_template = """
                    <div class="status">%message%<div class="time">%time%</div></div>
                    """;
                content_stylesheet = themed_style_sheet ();
            }

            message_parts = new GLib.List<MessagePart> ();
            var body_chunk = new StringBuilder ();
            foreach (var child in thread) {
                MessagePart? html_part = null;
                MessagePart? text_part = null;

                try {
                    if (child.uri == "error:")
                        throw new GLib.FileError.FAILED (child.subject);
                    child.parse_body ();
                    child.parse_text ();
                    html_part = child.html_part;
                    text_part = child.text_part;
                }
                catch (GLib.Error parse_error) {
                    html_part = null;
                    text_part = new MessagePart ("text/plain");
                    text_part.body.append (parse_error.message);
                }

                var reply_chunk = new StringBuilder ();
                message_parts.append (html_part ?? (text_part ?? child.parts.nth_data (0)));
                if (html_part != null)
                    reply_chunk.append (html_part.body.str);
                else if (text_part != null)
                    reply_chunk.append (render_plain_text (text_part.body.str, child.project));
                else
                    reply_chunk.append (child.parts.nth_data (0).body.str);

                string? sender_name = null;
                if (child.can_reply_personally) {
                    sender_name = Contact.name_from_string (child.sender);
                    if (sender_name != _("You")) {
                        reply_chunk.append_printf ("""
                            <span class="contact">%s</span>
                        """.printf (
                        linkify_icon (STOCK_CONTACT_NEW, _("Add contact"), "contact:%s:%s".printf (
                            sender_name, html_escape (Contact.address_from_string (child.sender))))));
                    }

                    reply_chunk.append ("""
                        <div class="actions">
                        <a href="mailto:%s" class="button">%s</a>
                        <a href="mailto:%s" class="button">%s</a></div>
                        """.printf (
                        html_escape (reply_uri (child, "Re: ")), _("Reply"),
                        html_escape (reply_uri (child, "Fw: ")), _("Forward")));
                }

                string reply_markup = reply_chunk.str;
                /* The extra class importantMessage is Postler-specific */
                if (child.priority)
                    reply_markup = "<span class=\"importantMessage\">%s</span>".printf (
                        reply_markup);
                if (sender_name == _("You"))
                    reply_markup = outgoing_template.replace ("%message%", reply_markup);
                else
                    reply_markup = content_template.replace ("%message%", reply_markup);
                reply_markup = reply_markup.replace ("\"%sender%\"", ("\"" + html_escape (child.sender) + "\""));
                /* Do inherit colors, to match Adium and Empathy */
                reply_markup = reply_markup.replace ("%sender%", linkify_address (child.sender, true));
                reply_markup = reply_markup.replace ("%time{%X}%", format_date (child.date));
                string avatar_uri = child.get_avatar_uri () ?? "Incoming/buddy_icon.png";
                reply_markup = reply_markup.replace ("%userIconPath%", avatar_uri);

                foreach (var attachment_part in child.parts) {
                    if (attachment_part != html_part && attachment_part != text_part
                     && !attachment_part.mime_type.has_prefix ("multipart/")) {
                        message_parts.append (attachment_part);
                        /* The extra class attachmentMessage is Postler-specific */
                        if (next_content_template != null) {
                            reply_markup = reply_markup.replace (
                                "<span id=\"insert\"></span>",
                                "<span class=\"attachmentMessage\">%s</span>".printf (
                                    next_content_template));
                            reply_markup = reply_markup.replace ("%sender%", "");
                            reply_markup = reply_markup.replace ("%time{%X}%", "");
                        }
                        else {
                            reply_markup += "<span class=\"attachmentMessage\">%s</span>".printf (
                                status_template);
                            reply_markup = reply_markup.replace ("%time%", "");
                        }
                        reply_markup = reply_markup.replace ("%message%",
                            """
                            <b>%s</b><div class="actions">
                            <a href="message-part:save:%d" class="button">%s</a>
                            <a href="message-part:open:%d" class="button">%s</a></div>
                            """.printf (
                            attachment_part.filename ?? attachment_part.mime_type,
                            child.parts.position (child.parts.find (attachment_part)), _("Save"),
                            child.parts.position (child.parts.find (attachment_part)), _("Open")));
                    }
                }

                body_chunk.append (reply_markup);
            }

            /* Don't change the message if it's the same to avoid notifications */
            if (message == null || message.id != the_message.id)
                message = thread.nth_data (0);
            last_location = message.get_path ();
            subject = message.subject;
            /* If there's no real address, it's probably undisclosed-recipients:; */
            string recipients;
            if (message.recipients.chr (-1, '@') != null)
                recipients = linkify_address (reply_uri (message, "Re: ", message.recipients));
            else
                recipients = _("Undisclosed recipients");
            list_post = message.get_field ("list-post");
            if (list_post != null)
                list_post = Postler.Contact.address_from_string (list_post);
            list_unsubscribe = message.get_field ("list-unsubscribe");
            if (list_unsubscribe != null) {
                list_unsubscribe = Postler.Contact.address_from_string (list_unsubscribe);
                list_unsubscribe = linkify_address ("<%s> %s".printf (_("Unsubscribe"), list_unsubscribe));
            }

            /* Emoticons */
            string body_markup = body_chunk.str;
            /* foreach (var emoticon in emoticons) { */
            for (int i = 0; i < emoticons.length; i++) {
                var emoticon = emoticons[i];
                try {
                    var escaped = GLib.Regex.escape_string (" " + emoticon.token);
                    var regex = new GLib.Regex (escaped);
                    body_markup = regex.replace_eval (body_markup, -1, 0, 0,
                        (match_info, result) => { return evaluate_emoticon (match_info, result, emoticon); });
                }
                catch (GLib.RegexError regex_error) { }
            }

            load_string (
                """
                <style text="text/css">%s %s</style>
                <div class="headers">
                <b>%s</b><br>
                %s
                %s
                </div>
                <div class="body">%s</div>
                """
                .printf (
                         style_sheet.replace ("ButtonText",
                             color_to_rgb (style.fg[widget_state])),
                         content_stylesheet,
                         message.subject,
                         recipients,
                         format_header (_("Mailing List:"), list_unsubscribe),
                         body_markup),
                "text/html", "UTF-8", "file://" + template_path);
        } catch (GLib.Error error) {
            display_error (error.message);
        }
        return false;
    }

    string render_plain_text (string body, string? project) {
        bool in_signature = false;
        int in_quote = 0;

        var new_body = new StringBuilder ("");
        if (project != null)
            new_body.append ("<span class=\"plain_text no_signature\">");
        foreach (string line in body.split ("<br>")) {
            /* Looks like a signature */
            if (in_signature && line[0] == '\0') {
                new_body.append ("</span>");
                in_signature = false;
            } else if (!in_signature && line[0] == '-' && line[1] == '-'
                    && line[2] == ' ' && line[3] == '\0') {
                new_body.append ("<span class=\"signature\">");
                in_signature = true;
                line = "";
            }
            /* Looks like quoting */
            if (in_quote > 0 || line.has_prefix ("> ") || line.has_prefix (">>")) {
                /* Determine level of nesting */
                int quote_mark = 0;
                int position = 0;
                do {
                    if (line[position] == '>')
                        quote_mark++;
                    else if (line[position] != ' ')
                        break;
                    position++;
                } while (true);

                if (quote_mark < in_quote)
                    new_body.append ("</blockquote>");
                else if (quote_mark > in_quote) {
                    for (int i = in_quote; i < quote_mark; i++)
                        new_body.append ("<blockquote>");
                }
                in_quote = quote_mark;
            }

            if (in_quote > 0 && line[0] == '>' && line[1] == ' ') {
                line = line.substring (2);
                int position = 0;
                do {
                    if (line[position] != '>' && line[position] != ' ')
                        break;
                    position++;
                } while (true);
                line = line.substring (position);
            }
            else if (in_quote > 0 && line[0] == '>' && line[1] == '\0')
                line = line.substring (1);
            new_body.append (line + "<br>");
        }
        if (project != null)
            new_body.append ("</span");

        /* Linkify */
        try {
            foreach (var link_format in link_formats) {
                var regex = new GLib.Regex (link_format);
                new_body.str = regex.replace (new_body.str, -1, 0,
                    "<a href=\"\\1\">\\1</a>");
            }
        }
        catch (GLib.RegexError regex_error) { }

        return new_body.str;
    }

    string color_to_rgb (Gdk.Color? color) {
        if (color == null)
            return "inherit";
        return color.to_string ().substring (0, 7);
    }

#if HAVE_GTK3
    const Gtk.StateFlags widget_state = Gtk.StateFlags.NORMAL;
#else
    const Gtk.StateType widget_state = Gtk.StateType.NORMAL;
#endif

    string themed_style_sheet () {
        /* Retrieve style from toplevel in case widget's not visible yet */
        var widget = get_toplevel ();
        /* Use void* with style_get to avoid Vala compiler error */
        void* link_color;
        widget.style_get ("link-color", out link_color);
        return fallback_style_sheet.replace (
            "ButtonText", color_to_rgb (widget.style.fg[widget_state])).replace (
            "ButtonFace", color_to_rgb (widget.style.bg[widget_state])).replace (
            "WindowText", color_to_rgb (widget.style.fg[widget_state])).replace (
            "Window", color_to_rgb (widget.style.base[widget_state])).replace (
            "Link", color_to_rgb (*((Gdk.Color*)link_color)));
    }

    void display_error (string message) {
        load_string ("""
            <style text="text/css">%s</style>
            <body>
            <h1>%s</h1>
            <p>%s</p>
            </body>
            """.
            printf (themed_style_sheet (), _("Error"), message),
            "text/html", "UTF-8", "about:blank");
        GLib.critical (message);
    }

    bool navigation_decision (WebKit.WebFrame web_frame,
        WebKit.NetworkRequest request, WebKit.WebNavigationAction action,
        WebKit.WebPolicyDecision decision) {
        string? uri = request.get_uri ();

        if (uri == null || uri == "about:blank") {
            decision.use ();
            return true;
        }

        if (uri != null && uri.has_prefix ("about:"))
            return false;
        if (uri.has_prefix ("file:///"))
            return false;

        decision.ignore ();
        if (uri.has_prefix ("contact:")) {
            string[] parts = Soup.URI.decode (uri).split (":");
            new Dexter.Dexter ().edit_contact (parts[1], parts[2]);
            return true;
        }
        else if (uri.has_prefix ("message-part:")) {
            int index;
            if (uri.has_prefix ("message-part:open:"))
                index = uri.substring (18, -1).to_int ();
            else if (uri.has_prefix ("message-part:save:"))
                index = uri.substring (18, -1).to_int ();
            else {
                /* FIXME: Show error dialogue */
                GLib.message (_("Erroneous attachment: %s"), uri);
                return false;
            }
            var message_part = message_parts.nth_data (index);
            string filename;
            bool should_open = true;
            if (uri.has_prefix ("message-part:open:")) {
                filename = "/tmp/%sXXXXXX".printf (message_part.filename ?? "");
                if (FileUtils.mkstemp (filename) == -1) {
                    /* FIXME: Show error dialogue */
                    GLib.message (_("Failed to create temporary file: %s"), filename);
                    return false;
                }
            } else if (uri.has_prefix ("message-part:save:")) {
                var filechooser = new Gtk.FileChooserDialog (_("Save As..."),
                    null, Gtk.FileChooserAction.SAVE,
                    Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                    Gtk.STOCK_SAVE, Gtk.ResponseType.ACCEPT);
                filechooser.do_overwrite_confirmation = true;
                if (save_folder == null) {
                    save_folder = Environment.get_user_special_dir (UserDirectory.DOWNLOAD);
                    if (save_folder ==null)
                        save_folder = Environment.get_home_dir ();
                }
                filechooser.set_current_folder (save_folder);
                filechooser.set_current_name (message_part.filename ?? "");
                if (filechooser.run () == Gtk.ResponseType.ACCEPT) {
                    filename = filechooser.get_filename ();
                    save_folder = filechooser.get_current_folder_file ().get_path ();
                }
                else
                    filename = null;
                filechooser.destroy ();
                if (filename == null)
                    return true;
                should_open = false;
            } else
                assert_not_reached ();
            try {
                if (mime_type_is_text (message_part.mime_type))
                    FileUtils.set_contents (filename, message_part.body.str);
                else
                    FileUtils.set_data (filename,
                        GLib.Base64.decode (message_part.body.str));
                uri = Filename.to_uri (filename, null);
            } catch (GLib.Error error) {
                /* FIXME: Show error dialogue */
                GLib.message (_("Failed to write file: %s"), filename);
                return false;
            }
            if (!should_open)
                return true;
        }

        return Postler.App.show_uri (get_screen (), uri);
    }

    WebKit.WebView new_window_created (WebKit.WebFrame web_frame) {
        /* Forward new windows, we never open new windows of our own */
        return this;
    }

    public override bool scroll_event (Gdk.EventScroll event) {
        if ((event.state & Gdk.ModifierType.CONTROL_MASK) != 0) {
            if (event.direction == Gdk.ScrollDirection.UP) {
                zoom_in ();
                return true;
            } else if (event.direction == Gdk.ScrollDirection.DOWN) {
                zoom_out ();
                return true;
            }
        }
        return false;
    }
}

