/*
 Copyright (C) 2011 Christian Dywan <christian@twotoasts.de>

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 See the file COPYING for the full license text.
*/

namespace Postler {
    public class Index : GLib.Object, GLib.Initable {
        static Sqlite.Database? database = null;
        static Sqlite.Statement? statement_insert = null;
        static Sqlite.Statement? statement_insert_sender = null;
        static Sqlite.Statement? statement_get = null;
        static Sqlite.Statement? statement_thread = null;
        static Sqlite.Statement? statement_list = null;
        static Sqlite.Statement? statement_position = null;
        static Sqlite.Statement? statement_count = null;
        static Sqlite.Statement? statement_remove = null;
        static Sqlite.Statement? statement_unread = null;
        static Sqlite.Statement? statement_guess = null;
        static int64 startup_timestamp = new DateTime.now_utc ().to_unix ();
        static GLib.HashTable<string,Postler.Contact> names;
#if HAVE_FOLKS
        static Folks.IndividualAggregator aggregator;

        void populate_display_names () {
            foreach (var individual in aggregator.get_individuals ().get_values ()) {
                string? display_name = individual.full_name ?? individual.alias;
                if (display_name != null) {
                    var contact = new Postler.Contact (display_name, individual.avatar);
                    foreach (string address in individual.email_addresses)
                        names.insert (address, contact);
                    /* foreach (string address in individual.im_addresses.get_values ())
                            names.insert (address, contact); */
                    foreach (var persona in individual.personas) {
                        if (persona.display_id.chr (-1, '@') != null)
                            names.insert (persona.display_id, contact);
                    }
                }
            }
        }
#endif

        bool address_is_user (string address) {
            var accounts = new Accounts ();
            foreach (var account_info in accounts.get_infos ()) {
                if (account_info.type == AccountType.SEARCH)
                    continue;
                foreach (string the_address in account_info.address.split (",")) {
                    if (the_address == address)
                        return true;
                }
            }
            return false;
        }

        bool row_or_done (int result) {
            return result == Sqlite.ROW || result == Sqlite.DONE;
        }

        bool init (GLib.Cancellable? cancellable = null) throws GLib.Error {
            return true;
        }

        public Index () throws GLib.Error {
            if (database != null)
                return;

            names = new GLib.HashTable<string,Postler.Contact> (str_hash, str_equal);
            /* FIXME: Populate addresses from Dexter */
#if HAVE_FOLKS
            aggregator = new Folks.IndividualAggregator ();
            /* FIXME: Connect to aggregator.individuals_changed */
            aggregator.prepare ();
            aggregator.notify["user"].connect ((object, pspec) => {
                populate_display_names ();
            });
#endif

            unowned string data_dir = Environment.get_user_data_dir ();
            string data_path = data_dir + "/" + Config.PACKAGE_NAME + "/mail/";
            if (Sqlite.Database.open (data_path + "index.db", out database) != Sqlite.OK)
                throw new GLib.FileError.FAILED (_("Failed to open database: %s"), database.errmsg ());
            if (database.exec ("PRAGMA journal_mode = TRUNCATE;") != Sqlite.OK)
                throw new GLib.FileError.FAILED (_("Failed to open database: %s"), database.errmsg ());
            if (database.exec ("PRAGMA synchronous = OFF;") != Sqlite.OK)
                throw new GLib.FileError.FAILED (_("Failed to open database: %s"), database.errmsg ());
            if (database.exec (
                """
                CREATE TABLE IF NOT EXISTS
                messages (id TEXT, uri TEXT, threads TEXT, project TEXT, list TEXT,
                          subject TEXT, sender TEXT, recipients TEXT,
                          unread BOOLEAN, flagged BOOLEAN,
                          forwarded BOOLEAN, replied BOOLEAN, priority BOOLEAN,
                          date INTEGER, timestamp INTEGER, excerpt TEXT);
                """
                ) != Sqlite.OK)
                throw new GLib.FileError.FAILED (_("Failed to open database: %s"), database.errmsg ());

            string[] indices = {
                "CREATE INDEX IF NOT EXISTS id_index ON messages (id);",
                "CREATE INDEX IF NOT EXISTS threads_index ON messages (threads);",
                "CREATE TABLE IF NOT EXISTS senders (address TEXT, name TEXT)",
                "CREATE INDEX IF NOT EXISTS senders_index ON senders (address);"
            };
            foreach (string index_statement in indices) {
                if (database.exec (index_statement) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to open database: %s"), database.errmsg ());
            }
        }

        public void insert_sender (string address) throws GLib.Error {
            if (statement_insert_sender == null) {
                if (database.prepare_v2 ("""
                    INSERT INTO senders (address, name) VALUES (?1, ?2)
                    """,
                    -1, out statement_insert_sender) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to index message: %s"), database.errmsg ());
            }
            string[] parts = Contact.parse (address);
            /* Don't collect address <address> which is not useful */
            if (parts[0] == parts[1])
                return;
            bool success =
                statement_insert_sender.bind_text (1, parts[1], -1) == Sqlite.OK
             && statement_insert_sender.bind_text (2, parts[0], -1) == Sqlite.OK
             && statement_insert_sender.step () == Sqlite.DONE;
            statement_insert_sender.reset ();
            if (!success)
                throw new GLib.FileError.FAILED (_("Failed to index message: %s"), database.errmsg ());
        }

        public void insert_received_message (Message message) throws GLib.Error {
            if (statement_insert == null) {
                if (database.prepare_v2 ("""
                    INSERT INTO messages
                    (uri, subject, sender, recipients,
                     unread, flagged, forwarded, replied, priority,
                     date, id, timestamp, threads, project, list, excerpt)
                    VALUES (?1, ?2, ?3, ?4, ?5, ?6, ?7, ?8, ?9, ?10, ?11, ?12, ?13, ?14, ?15, ?16)
                    """,
                    -1, out statement_insert) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to index message: %s"), database.errmsg ());
            }
            bool success =
                statement_insert.bind_text (1, message.uri, -1) == Sqlite.OK
             && statement_insert.bind_text (2, message.subject, -1) == Sqlite.OK
             && statement_insert.bind_text (3, message.sender, -1) == Sqlite.OK
             && statement_insert.bind_text (4, message.recipients, -1) == Sqlite.OK
             && statement_insert.bind_int64 (5, message.unread ? 1 : 0) == Sqlite.OK
             && statement_insert.bind_int64 (6, message.flagged ? 1 : 0) == Sqlite.OK
             && statement_insert.bind_int64 (7, message.forwarded ? 1 : 0) == Sqlite.OK
             && statement_insert.bind_int64 (8, message.replied ? 1 : 0) == Sqlite.OK
             && statement_insert.bind_int64 (9, message.priority ? 1 : 0) == Sqlite.OK
             && statement_insert.bind_int64 (10, message.get_timestamp ()) == Sqlite.OK
             && statement_insert.bind_text (11, message.id) == Sqlite.OK
             && statement_insert.bind_int64 (12, startup_timestamp) == Sqlite.OK
             && statement_insert.bind_text (13, message.threads, -1) == Sqlite.OK
             && statement_insert.bind_text (14, message.project, -1) == Sqlite.OK
             && statement_insert.bind_text (15, message.list, -1) == Sqlite.OK
             && statement_insert.bind_text (16, message.excerpt, -1) == Sqlite.OK
             && statement_insert.step () == Sqlite.DONE;
            statement_insert.reset ();
            if (!success)
                throw new GLib.FileError.FAILED (_("Failed to index message: %s"), database.errmsg ());
            insert_sender (message.sender);
#if HAVE_ZEITGEIST
            var log = new Zeitgeist.Log ();
            var event = new Zeitgeist.Event.full (Zeitgeist.ZG_RECEIVE_EVENT,
                Zeitgeist.ZG_SYSTEM_NOTIFICATION, "app://postler.desktop",
                new Zeitgeist.Subject.full (message.uri,
                Zeitgeist.NMO_EMAIL, Zeitgeist.NMO_MAILBOX_DATA_OBJECT,
                "text/plain", GLib.File.new_for_uri (message.uri).get_parent ().get_uri (),
                message.sender, ""));
            log.insert_events_no_reply (event);
#endif
        }

        public bool update_filename (string uri) throws GLib.Error {
            if (statement_count == null) {
                if (database.prepare_v2 ("""
                    UPDATE messages SET timestamp = ?2 WHERE uri = ?1
                    """,
                    -1, out statement_count) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to update filename: %s"), database.errmsg ());
            }
            bool success =
                statement_count.bind_text (1, uri, -1) == Sqlite.OK
             && statement_count.bind_int64 (2, startup_timestamp) == Sqlite.OK
             && statement_count.step () == Sqlite.DONE;
            if (!success) {
                statement_count.reset ();
                throw new GLib.FileError.FAILED (_("Failed to update filename: %s"), database.errmsg ());
            }
            int64 updated_rows = database.changes ();
            statement_count.reset ();
            return updated_rows > 0;
        }

        public void purge_filenames () throws GLib.Error {
            if (database.exec (
                "DELETE FROM messages WHERE timestamp < " + startup_timestamp.to_string ())
                != Sqlite.OK)
                throw new GLib.FileError.FAILED (_("Failed to purge old filenames: %s"), database.errmsg ());
            startup_timestamp = new DateTime.now_utc ().to_unix ();
        }

        public void remove_filename (string uri) throws GLib.Error {
            if (statement_remove == null) {
                if (database.prepare_v2 ("""
                    DELETE FROM messages WHERE uri = ?1
                    """,
                    -1, out statement_remove) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to remove filename: %s"), database.errmsg ());
            }
            bool success =
                statement_remove.bind_text (1, uri, -1) == Sqlite.OK
             && statement_remove.step () == Sqlite.DONE;
            if (!success) {
                statement_remove.reset ();
                throw new GLib.FileError.FAILED (_("Failed to remove filename: %s"), database.errmsg ());
            }
            statement_remove.reset ();
        }

        string search_statement (string folder, string sql,
            ref Sqlite.Statement statement) throws GLib.Error {

            /* file:// or search:field/value */
            if (folder.has_prefix ("search:")) {
                string[] parts = folder.slice (7, -1).split ("/", 2);
                if (parts[0] == null || parts[1] == null)
                    throw new GLib.FileError.FAILED (_("Failed to list messages: %s"),
                                                     "parts[0] == null || parts[1] == null");
                string field = parts[0];
                if (field == "from")
                    field = "sender";
                else if (field == "to")
                    field = "recipients";
                else if (field == "fulltext")
                    field = "excerpt";
                /* We can use the prepared statement for uri search */
                if (field != "uri") {
                    /* Filter out Trash and Junk */
                    string real_sql = sql.replace ("$CASE",
                        "AND uri NOT LIKE '%/Trash/%' AND uri NOT LIKE '%/Junk/%'");
                    if (database.prepare_v2 (real_sql.replace ("$FIELD", field),
                                             -1, out statement) != Sqlite.OK)
                        throw new GLib.FileError.FAILED (_("Failed to list messages: %s"), database.errmsg ());
                }
                return "%%%s%%".printf (parts[1]);
            }
            return folder;
        }

        public int64[] unread_messages (string folder) throws GLib.Error {
            unowned string sql = """
                SELECT unread, COUNT (*) FROM messages
                WHERE $FIELD LIKE ?2 $CASE GROUP BY unread
                """;

            if (statement_unread == null) {
                if (database.prepare_v2 (sql.replace ("$FIELD", "uri").replace ("$CASE", ""),
                                         -1, out statement_unread) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to count unread messages: %s"), database.errmsg ());
            }
            Sqlite.Statement? temporary_statement = null;
            string search_value = search_statement (folder, sql, ref temporary_statement);
            unowned Sqlite.Statement statement = temporary_statement != null
                                               ? temporary_statement : statement_unread;
            int result = Sqlite.ERROR;
            bool success = statement.bind_text (2, search_value, -1) == Sqlite.OK
             && (row_or_done ((result = statement.step ())));
            if (!success) {
                statement.reset ();
                throw new GLib.FileError.FAILED (_("Failed to count unread messages: %s"), database.errmsg ());
            }
            int64 read = 0, unread = 0;
            while (result == Sqlite.ROW) {
                if (statement.column_int64 (0) == 0)
                    read = statement.column_int64 (1);
                else
                    unread = statement.column_int64 (1);
                result = statement.step ();
            }
            statement.reset ();
            return { unread, unread + read };
        }

        public Message get_message (string id) throws GLib.Error {
            if (statement_get == null) {
                if (database.prepare_v2 ("""
                    SELECT subject, sender, 0,
                    unread, flagged, forwarded, replied, priority, date, uri, excerpt,
                    (SELECT COUNT (*) FROM messages WHERE threads = ?1 AND unread = 1) AS replies
                    FROM messages
                    WHERE id = ?1 GROUP BY id ORDER BY date ASC LIMIT 1
                    """,
                    -1, out statement_get) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to list messages: %s"), database.errmsg ());
            }
            bool success =
                statement_get.bind_text (1, id, -1) == Sqlite.OK
             && row_or_done (statement_get.step ());
            if (!success) {
                statement_get.reset ();
                throw new GLib.FileError.FAILED (_("Failed to get message: %s"), database.errmsg ());
            }
            var message = new Message.from_statement (id, statement_get);
            string address = Postler.Contact.address_from_string (message.sender);
            try {
                var contact = guess_name (address);
                if (contact != null) {
                    message.sender = "\"%s\" <%s>".printf (contact.display_name, address);
                    message.avatar = contact.avatar;
                }
            }
            catch (GLib.Error guess_error) {
                GLib.warning (guess_error.message);
            }
            statement_get.reset ();
            return message;
        }

        public string[] get_thread (string thread) throws GLib.Error {
            if (statement_thread == null) {
                if (database.prepare_v2 ("""
                    SELECT id FROM messages WHERE threads = ?1 GROUP BY id ORDER BY date ASC
                    """,
                    -1, out statement_thread) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to thread messages: %s"), database.errmsg ());
            }
            int result = Sqlite.ERROR;
            bool success =
                statement_thread.bind_text (1, thread, -1) == Sqlite.OK
             && row_or_done (((result = statement_thread.step ())));
            if (!success) {
                statement_thread.reset ();
                throw new GLib.FileError.FAILED (_("Failed to thread messages: %s"), database.errmsg ());
            }
            string[] ids = {};
            while (result == Sqlite.ROW) {
                ids += statement_thread.column_text (0);
                result = statement_thread.step ();
            }
            statement_thread.reset ();
            return ids;
        }

        public string[] get_messages (string folder) throws GLib.Error {
            unowned string sql = """
                SELECT threads AS thread_id FROM messages
                WHERE $FIELD LIKE ?2 $CASE GROUP BY threads
                HAVING (SELECT COUNT (*) FROM messages WHERE id = thread_id)
                ORDER BY date ASC
                """;

            if (statement_list == null) {
                if (database.prepare_v2 (sql.replace ("$FIELD", "uri").replace ("$CASE", ""),
                                         -1, out statement_list) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to list messages: %s"), database.errmsg ());
            }

            Sqlite.Statement? temporary_statement = null;
            string search_value = search_statement (folder, sql, ref temporary_statement);
            unowned Sqlite.Statement statement = temporary_statement != null
                                               ? temporary_statement : statement_list;
            bool success = statement.bind_text (2, search_value, -1) == Sqlite.OK;
            int result = Sqlite.ERROR;
            success = success && row_or_done (((result = statement.step ())));
            if (!success) {
                statement.reset ();
                throw new GLib.FileError.FAILED (_("Failed to list messages: %s"), database.errmsg ());
            }
            string[] ids = {};
            while (result == Sqlite.ROW) {
                ids += statement.column_text (0);
                result = statement.step ();
            }
            statement.reset ();
            return ids;
        }

        public int64 get_position (Message message) throws GLib.Error {
            if (statement_position == null) {
                if (database.prepare_v2 ("""
                    SELECT id FROM messages WHERE uri LIKE ?1 GROUP BY id ORDER BY date ASC
                    """,
                    -1, out statement_position) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to locate message: %s"), database.errmsg ());
            }
            int result = Sqlite.ERROR;
            bool success =
                statement_position.bind_text (1, message.uri, -1) == Sqlite.OK
             && row_or_done (((result = statement_position.step ())));
            if (!success) {
                statement_position.reset ();
                throw new GLib.FileError.FAILED (_("Failed to locate message: %s"), database.errmsg ());
            }
            int64 position = -1;
            while (result == Sqlite.ROW) {
                position++;
                if (message.id == statement_position.column_text (0)) {
                    statement_position.reset ();
                    return position;
                }
                result = statement_position.step ();
            }
            statement_position.reset ();
            return -1;
        }

        public Postler.Contact? guess_name (string address) throws GLib.Error {
            Postler.Contact? contact = names.lookup (address);
            if (contact != null) {
                if (address_is_user (address))
        /* i18n: 'You' refers to the user, displayed next to messages the user sent */
                    return new Postler.Contact (_("You"), contact.avatar);
                return contact;
            }

            if (address_is_user (address))
                return new Postler.Contact (_("You"), null);

            if (statement_guess == null) {
                if (database.prepare_v2 ("""
                    SELECT name FROM senders WHERE address = ?1
                    GROUP BY name ORDER BY COUNT (name) DESC LIMIT 1
                    """,
                    -1, out statement_guess) != Sqlite.OK)
                    throw new GLib.FileError.FAILED (_("Failed to guess name: %s"), database.errmsg ());
            }
            int result = Sqlite.ERROR;
            bool success =
                statement_guess.bind_text (1, address, -1) == Sqlite.OK
             && row_or_done (((result = statement_guess.step ())));
            if (!success) {
                statement_guess.reset ();
                throw new GLib.FileError.FAILED (_("Failed to guess name: %s"), database.errmsg ());
            }
            if (result == Sqlite.ROW) {
                string display_name = statement_guess.column_text (0);
                statement_guess.reset ();
                return new Postler.Contact (display_name, null);
            }
            statement_guess.reset ();
            return null;
        }
    }
}

