/*
 Copyright (C) 2011 Christian Dywan <christian@twotoasts.de>

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 See the file COPYING for the full license text.
*/

namespace Postler {
    public class MessagePart : Object {
        public StringBuilder? body;
        public string? mime_type;
        public string? charset;
        public bool plain_text;
        public string? content_disposition;
        public string? content_id;
        public string? filename;

        public MessagePart (string mime_type) {
            body = new StringBuilder ();
            this.mime_type = mime_type;
            plain_text = mime_type == "text/plain";
            filename = null;
        }

        public bool is_empty () {
            bool empty = true;
            if (!(body == null)) {
                var b_str = body.str.strip ();
                string[] b_str_parts = b_str.split ("<br>");
                long str_length = 0;
                foreach (var current_line in b_str_parts) {
                    string[] split_split = current_line.strip().split("\n");
                    foreach (var between_linebreaks in split_split)
                    str_length = between_linebreaks.strip().length;
                }
                if (str_length != 0)
                    empty = false;
            }
            return empty;
        }

        public string get_plain_text () {
            string body_plain = body.str;
            if ((body != null) && (mime_type != null)) {
                try {
                    string plain_text = "";
                    int std_in;
                    int std_out;
                    string[] argv = { "lynx", "-force_html", "-width=900", "-dump",
                                      "-display_charset=utf-8", "-stdin" };
                    Process.spawn_async_with_pipes (null, argv, null,
                        SpawnFlags.SEARCH_PATH, null, null,
                        out std_in, out std_out, null);
                    IOChannel ioc_stdout = new IOChannel.unix_new (std_out);
                    IOChannel ioc_stdin = new IOChannel.unix_new (std_in);
                    char[] content_chars = (char[]) (body.str.data);
                    var status = ioc_stdin.write_chars (content_chars, null);
                    ioc_stdin.shutdown (true);
                    if (status != IOStatus.ERROR)
                        status = ioc_stdout.read_to_end (out plain_text, null);
                    if (status != IOStatus.ERROR)
                        body_plain = plain_text;
                }
                catch (Error error) {
                    GLib.message (_("Error converting HTML to text: %s"), error.message);
                }
            }
            return body_plain;
        }
    }

    public enum MessageFlags {
        NONE = '0',
        MARK_READ = '1',
        TOGGLE_READ = 'S',
        TOGGLE_FLAGGED = 'F',
        TOGGLE_FORWARDED = 'P',
        TOGGLE_REPLIED = 'R',
    }

    public class Message : GLib.Object, GLib.Initable {
        public string? uri { public get; set; }
        public string get_path () { return GLib.File.new_for_uri (uri).get_path (); }
        public string? id { public get; set; }
        public string? threads { public get; set; }
        public int64 replies { public get; set; }
        string? charset = null;
        public string get_charset () { return charset ?? "ISO-8859-1"; }
        public GLib.DateTime? date { public get; set; }
        public int64 get_timestamp () { return date != null ? date.to_unix () : 0; }
        public string? subject { public get; set; }
        public string? sender { public get; internal set; }
        public GLib.File? avatar { get; internal set; }
        public string? recipients { public get; set; }
        public string? reply_to { public get; set; }
        public bool can_reply_personally { public get; set; default = true; }
        public bool unread { public set; get; }
        public bool flagged { public get; set; }
        public bool forwarded { public get; set; }
        public bool replied { public get; set; }
        public bool priority { public get; set; }
        public string? excerpt { public get; set; }
        public string organization { public get; set; }
        public string application { public get; set; }
        public string project { public get; set; }
        public string list { public get; set; }
        GLib.HashTable<string,string> fields = new GLib.HashTable<string,string> (str_hash, str_equal);
        public string get_field (string field) { return fields.lookup (field); }
        public GLib.List<MessagePart>? parts = null;
        public GLib.DataInputStream? stream = null;
        public MessagePart? html_part = null;
        public MessagePart? text_part = null;

        public string get_avatar_uri () {
            if (avatar != null && avatar.get_uri () != null)
                return avatar.get_uri ();
            string sender_checksum = GLib.Checksum.compute_for_string (
                GLib.ChecksumType.MD5, Contact.address_from_string (sender));
            return "http://www.gravatar.com/avatar/%s?d=identicon".printf (sender_checksum);
        }

        bool init (GLib.Cancellable? cancellable = null) throws GLib.Error {
            return false;
        }

        static string parse_encoded (string quoted, out string charset) {
            return Postler.Messages.parse_encoded (quoted, out charset);
        }

        public GLib.HashTable<string,Variant> to_hash_table () {
            var message_data = new GLib.HashTable<string,Variant> (str_hash, str_equal);
            message_data.insert ("id", this.id);
            message_data.insert ("uri", this.uri ?? "");
            message_data.insert ("replies", this.replies);
            /* Enforce UTF-8 here, GLib.Variant otherwise crashes */
            message_data.insert ("subject", Postler.Messages.fallback_to_utf8 (subject ?? ""));
            message_data.insert ("sender", this.sender ?? "");
            if (avatar != null)
                message_data.insert ("avatar", avatar.get_uri ());
            message_data.insert ("unread", this.unread);
            message_data.insert ("flagged", this.flagged);
            message_data.insert ("forwarded", this.forwarded);
            message_data.insert ("replied", this.replied);
            message_data.insert ("priority", this.priority);
            message_data.insert ("excerpt", this.excerpt ?? "");
            message_data.insert ("date", this.get_timestamp ());
            return message_data;
        }

        public Message.from_id (string id) {
            this.id = id;
        }

        public void from_error (string error_message) {
            this.uri = "error:";
            this.subject = error_message;
        }

        public void from_hash_table (GLib.HashTable<string,Variant> message_data) {
            this.uri = message_data.lookup ("uri").get_string ();
            this.replies = message_data.lookup ("replies").get_int64 ();
            this.subject = message_data.lookup ("subject").get_string ();
            this.sender = message_data.lookup ("sender").get_string ();
            Variant? avatar = message_data.lookup ("avatar");
            if (avatar != null)
                this.avatar = GLib.File.new_for_uri (avatar.get_string ());
            this.unread = message_data.lookup ("unread").get_boolean ();
            this.flagged = message_data.lookup ("flagged").get_boolean ();
            this.forwarded = message_data.lookup ("forwarded").get_boolean ();
            this.replied = message_data.lookup ("replied").get_boolean ();
            this.priority = message_data.lookup ("priority").get_boolean ();
            this.excerpt = message_data.lookup ("excerpt").get_string ();
            this.date = new GLib.DateTime.from_unix_utc (message_data.lookup ("date").get_int64 ());
        }

        public Message.from_statement (string id, Sqlite.Statement statement) {
            this.id = id;
            this.replies = statement.column_int64 (11);
            this.subject = statement.column_text (0);
            this.sender = statement.column_text (1);
            this.unread = statement.column_int64 (3) != 0;
            this.flagged = statement.column_int64 (4) != 0;
            this.forwarded = statement.column_int64 (5) != 0;
            this.replied = statement.column_int64 (6) != 0;
            this.priority = statement.column_int64 (7) != 0;
            this.date = new GLib.DateTime.from_unix_utc (statement.column_int64 (8));
            this.uri = statement.column_text (9);
            this.excerpt = statement.column_text (10);
        }

        void read_flags (string path) {
            /* format "unique:2,DFPRST", ordered alphabetically */
            unowned string? last_colon = path.rstr (":2,");
            if (last_colon != null) {
                this.unread = last_colon.chr (-1, MessageFlags.TOGGLE_READ) == null;
                this.flagged = last_colon.chr (-1, MessageFlags.TOGGLE_FLAGGED) != null;
                this.forwarded = last_colon.chr (-1, MessageFlags.TOGGLE_FORWARDED) != null;
                this.replied = last_colon.chr (-1, MessageFlags.TOGGLE_REPLIED) != null;
            }
            else {
                this.unread = true;
            }
        }

        public void toggle_flag (char flag) {
            if (flag == MessageFlags.NONE)
                return;
            if (flag == MessageFlags.MARK_READ && unread)
                flag = MessageFlags.TOGGLE_READ;
            else if (flag == MessageFlags.MARK_READ)
                return;
            /* The message itself counts as a reply even as a parent */
            if (flag == MessageFlags.TOGGLE_READ)
                replies += unread ? -1 : +1;
            string old_path = get_path ();
            return_if_fail (old_path != null);
            string new_path = Postler.Messages.toggle_flag (old_path, flag);
            if (FileUtils.rename (old_path, new_path) == 0) {
                uri = "file://" + new_path;
                read_flags (new_path);
            }
        }

        public void move (string? destination) throws GLib.Error {
            var file = File.new_for_uri (uri);
            if (destination == null) {
                file.delete ();
                return;
            }

            string new_location = Postler.Messages.update_filename (
                file.get_path (), destination + "/cur/");
            var destination_file = File.new_for_path (new_location);
            file.move (destination_file, GLib.FileCopyFlags.NONE);
        }

        public AccountInfo? get_account (Accounts accounts) {
            if (!(uri != null && uri.has_prefix ("file://")))
                return null;

            /* Look for the account belonging to the message */
            string folder = uri.substring (0, uri.pointer_to_offset (uri.rstr ("/")));
            return_val_if_fail (folder.has_suffix ("/new") || folder.has_suffix ("/cur"), null);
            folder = folder.substring (0, folder.pointer_to_offset (folder.rstr ("/")));
            folder = folder.substring (0, folder.pointer_to_offset (folder.rstr ("/")));
            folder = folder.substring (7, -1);
            foreach (unowned AccountInfo info in accounts.get_infos ()) {
                if (info.type != AccountType.SEARCH && info.path.has_prefix (folder))
                    return info;
            }
            return new AccountInfo ();
        }

        public Message.from_file (GLib.File file,
            GLib.Cancellable? cancellable = null) throws GLib.Error {
            read_from_file (file, cancellable);
        }

        void read_from_file (GLib.File file,
            GLib.Cancellable? cancellable = null) throws GLib.Error {
            uri = file.get_uri ();
            read_flags (file.get_path ());
            read_from_stream (new DataInputStream (file.read (cancellable)));
        }

        public Message.from_stream (GLib.DataInputStream stream,
            GLib.Cancellable? cancellable = null) throws GLib.Error {
            uri = "memory:";
            read_from_stream (stream);
        }

        void read_from_stream (GLib.DataInputStream stream,
            GLib.Cancellable? cancellable = null) throws GLib.Error {

            this.recipients = null;
                
            this.stream = stream;
            string line = stream.read_line (null, cancellable);
            string next_line = "";
            while (line != null && line != "" && (next_line = stream.read_line (null, cancellable)) != null) {
                if (next_line[0] == '\t' || next_line[0] == ' ') {
                    line = line + " " + next_line.chug ();
                    continue;
                }

                string[] parts = line.split (":", 2);
                if (parts == null || parts[0] == null)
                    continue;

                string field = ascii_strdown (parts[0]);
                if (field == "message-id") {
                    string message_id = parts[1].strip ();
                    if (message_id.has_prefix ("<") && message_id.has_suffix (">"))
                        id = message_id.slice (1, -1);
                }
                else if (field == "in-reply-to") {
                    string message_id = parts[1].strip ();
                    if (message_id.has_prefix ("<") && message_id.has_suffix (">")) {
                        if (threads != null)
                            threads += "," + message_id.slice (1, -1);
                        else
                            threads = message_id.slice (1, -1);
                    }
                }
                /* else if (field == "references") {
                    string[] message_ids = parts[1].split_set (", \t");
                    string[] parsed_ids = {};
                    foreach (string message_id in message_ids) {
                        if (message_id.has_prefix ("<") && message_id.has_suffix (">"))
                            parsed_ids += message_id.slice (1, -1);
                    }
                    if (threads != null)
                        threads = string.joinv (",", parsed_ids) + "," + threads;
                    else
                        threads = string.joinv (",", parsed_ids);
                } */
                else if (field == "subject") {
                    subject = parse_encoded (parts[1], out charset);
                    if (subject.has_prefix ("Re: "))
                        subject = subject.substring (3);
                    /* FIXME: Parse a proper excerpt */
                    excerpt = subject;
                }
                else if (field == "from") {
                    string sender_charset = null;
                    fields.insert (field, parse_encoded (parts[1], out sender_charset));
                    if (sender == null)
                        sender = get_field (field);
                }
                else if (field == "x-bugzilla-who") {
                    string sender_charset = null;
                    if (get_field ("from") == sender)
                        sender = parse_encoded (parts[1], out sender_charset);
                }
                else if (field == "date") {
                    date = Postler.Content.date_from_string (parts[1]);}
                else if (field == "to" || field == "cc" || field == "bcc") {
                    recipients = (recipients ?? "") + parts[1].strip () + ",";
                    string field_charset = null;
                    fields.insert (field, parse_encoded (parts[1], out field_charset));
                }
                else if (field == "x-priority" || field == "importance")
                    priority = parts[1][0] == '1' || parts[1][0] == '2'
                            || parts[1] == "High";
                else if (field == "organization")
                    organization = parts[1];
                else if (field == "x-mailer" || field == "user-agent")
                    application = Postler.Content.format_x_mailer (parts[1]);
                else if (field == "x-bugzilla-product" || field == "x-launchpad-bug") {
                    /* Format:
                       X-Bugzilla-Product: gtk+
                       X-Launchpad-Bug: product=postler; status=Confirmed;
                       There can be multiple occurences of the Launchpad header. */
                    string stripped = parts[1].strip ();
                    bool semicolon = stripped.str ("; ") != null;
                    if (stripped.has_prefix ("product=") && semicolon)
                        project = stripped.split (";")[0].substring (8);
                    else if (!semicolon)
                        project = stripped;
                    /* Launchpad doesn't expose the reporter's address */
                    if (field == "x-launchpad-bug")
                        can_reply_personally = false;
                }
                else if (field == "list-id") {
                    string field_charset = null;
                    list = parse_encoded (parts[1], out field_charset);
                }
                else if (field == "reply-to") {
                    string field_charset = null;
                    reply_to = parse_encoded (parts[1], out field_charset);
                }
                else if (field == "list-post" || field == "list-unsubscribe") {
                    string field_charset = null;
                    fields.insert (field, parse_encoded (parts[1], out field_charset));
                }
                else if (field == "content-type" || field == "content-transfer-encoding")
                    fields.insert (field, parts[1]);

                line = next_line;
            }

            /* Treat top-levels like replies to themselves */
            if (threads == null)
                threads = id;

            recipients = parse_encoded (recipients, out charset);

            /* Some mailing list systems override Reply-To so that it doesn't
               point to the author but the list.
               Also Reply-To may equal From, which is at best confusing. */
            if (reply_to != null) {
                string canonical = Postler.Contact.address_from_string (reply_to);
                foreach(var r in Postler.Contact.quoted_split (recipients)) {
                    if (Postler.Contact.equal (canonical, r)) {
                        reply_to = null;
                        break;
                    }
                } 
                if (Postler.Contact.equal (canonical, get_field ("list-post"))) {
                    reply_to = null;
                }
                else if (Postler.Contact.equal (canonical, sender)) {
                    reply_to = null;
                }
            }

            if (id == null)
                throw new GLib.FileError.FAILED (_("Failed to find valid message ID"));
        }

        public static void parse_content_type (string? content_type, ref string charset,
            ref string boundary, ref string mime_type, ref string filename) {

            if (content_type == null)
                return;

            string[] parts = content_type.replace (" = ", "=").split_set ("; \t");
            filename = null;
            for (int i = 0; i < parts.length; i++) {
                string part = parts[i];
                unowned string? next = parts[i + 1];
                if ("\"" in part && next != null && "\"" in next) {
                    part += parts[i + 1];
                    i++;
                }

                if (part.has_prefix ("charset="))
                    charset = part.substring (8, part.length - 8).replace ("\"", "");
                else if (part.has_prefix ("name=") || part.has_prefix ("NAME="))
                    filename = part.substring (5, part.length - 5).replace ("\"", "");
                else if (part.has_prefix ("filename="))
                    filename = part.substring (9, part.length - 9).replace ("\"", "");
                else if (part.down ().has_prefix ("boundary="))
                    boundary = part.substring (9, part.length - 9).replace ("\"", "");
                else if (part != "" && !part.contains ("="))
                    mime_type = ascii_strdown (part);
            }

            if (filename != null) {
                string filename_charset;
                filename = parse_encoded (filename, out filename_charset);
            }

            if (mime_type == "application/octet-stream" && filename != null) {
                uchar[] data = {};
                bool uncertain;
                mime_type = g_content_type_guess (filename, data, out uncertain);
            }
        }

        bool mime_type_is_text (string mime_type) {
            return g_content_type_is_a (mime_type, "text/plain");
        }

        static bool evaluate_hex (GLib.MatchInfo match_info, GLib.StringBuilder result) {
            string match = "0x" + match_info.fetch (2);
            result.append_printf ("%c", (int)match.to_ulong (null));
            return false;
        }

        public static string quoted_printable_decode (string quoted) {
            /* =20 is not hex-encoded, to faciliate quote handling */
            string unquoted = quoted.replace ("=20", " ");
            try {
                var regex = new GLib.Regex ("([=]([0-9A-F][0-9A-F]))");
                return regex.replace_eval (unquoted, -1, 0, 0, evaluate_hex);
            }
            catch (GLib.RegexError error) {
                GLib.critical (_("Failed to decode string \"%s\": %s"),
                               unquoted, error.message);
            }
            return unquoted;
        }

        public void parse_body (GLib.Cancellable? cancellable=null) throws GLib.Error {
            if (stream == null) {
                return_if_fail (uri != null);
                read_from_file (GLib.File.new_for_uri (uri), cancellable);
            }

            parts = new GLib.List<MessagePart> ();
            string? content_encoding = get_field ("content-transfer-encoding");
            string? boundary = null;
            string mime_type = "text/plain";
            string? fname = null;
            parse_content_type (get_field ("content-type"), ref charset,
                                ref boundary, ref mime_type, ref fname);

            uint multipart = mime_type.has_prefix ("multipart/") ? 1 : 0;
            if (multipart > 0 && boundary != null)
                boundary = boundary.replace ("\"", " ").strip ();
            else
                multipart = 0;
            if (!(mime_type_is_text (mime_type) || multipart > 0)) {
                throw new GLib.FileError.FAILED (
                    _("This message of type \"%s\" can't be displayed.").
                    printf (mime_type));
            }

            MessagePart? message_part = null;
            /* Message body starts here */
            if (multipart == 0) {
                message_part = new MessagePart (mime_type);
                parts.append (message_part);
            }

            string inner_boundary = "";
            string line;
            string previous_line = "";
            while ((line = stream.read_line (null, null)) != null) {
                if (multipart > 0) {
                    if (line.has_prefix ("--")) {
                        if (line == "--" + inner_boundary) {
                            message_part = new MessagePart ("text/plain");
                            parts.append (message_part);
                            multipart = 2;
                            continue;
                        } else if (line == "--" + boundary) {
                            inner_boundary = "";
                            message_part = new MessagePart ("text/plain");
                            parts.append (message_part);
                            multipart = 2;
                            continue;
                        } else if (line == "--" + boundary + "--") {
                            /* Final boundary, we can skip it */
                            continue;
                        }
                    }
                    else if (multipart == 2) {
                        if (line == "") {
                            multipart++;
                            continue;
                        }

                        if (line[0] == '\t' || line[0] == ' ')
                            line = previous_line + " " + line.chug ();
                        previous_line = line;

                        string[] parts = line.split (":", 2);
                        string field = ascii_strdown (parts[0] ?? "");
                        if (field == "content-type") {
                            string ctype = get_charset ();
                            string mtype = "text/plain";
                            fname = null;
                            parse_content_type (parts[1].strip (), ref ctype,
                                ref inner_boundary, ref mtype, ref fname);
                            message_part.filename = fname;
                            message_part.mime_type = mtype;
                            message_part.charset = ctype;
                            message_part.plain_text = mtype == "text/plain";
                        }
                        else if (field == "content-transfer-encoding") {
                            content_encoding = ascii_strdown (parts[1].strip ());
                        }
                        else if (field == "content-disposition" ) {
                            string content_disposition = parts[1].strip ();
                            string cset = get_charset ();
                            string mtype = "text/plain";
                            fname = null;
                            parse_content_type (content_disposition, ref cset,
                                ref inner_boundary, ref mtype, ref fname);
                            message_part.filename = fname;
                            if (content_disposition.has_prefix ("attachment"))
                                message_part.content_disposition = content_disposition;
                            else if (content_disposition.has_prefix ("inline"))
                                message_part.content_disposition = content_disposition;
                        }
                        else if (field == "content-id") {
                            string content_id = parts[1].strip ();
                            if (content_id[0] == '<' && content_id.has_suffix (">"))
                                message_part.content_id = content_id.slice (1, -1);
                        }
                        continue;
                    }
                    else if (multipart == 1)
                        continue;
                }

                /* Compress subsequent empty lines */
                if (line == "" && previous_line == line)
                    continue;

                if (content_encoding == "quoted-printable")
                    line = quoted_printable_decode (line);
                else if (content_encoding == "base64"
                      && mime_type_is_text (message_part.mime_type))
                    line = line != "" ? (string)GLib.Base64.decode (line) : "";
                try {
                    string cset = message_part.charset ?? get_charset ();
                    if (cset != "UTF-8")
                        line = GLib.convert (line, -1, "UTF-8", cset, null);
                }
                catch (GLib.ConvertError error) { }
                if (message_part.plain_text)
                    line = line.replace ("<", "&lt;");
                if (content_encoding == "base64") {
                    if (message_part.plain_text)
                        line = line.replace ("\n", "<br>");
                    message_part.body.append (line);
                } else {
                    message_part.body.append (line);
                    if (content_encoding == "quoted-printable" && line.has_suffix ("="))
                        message_part.body.truncate (message_part.body.len - 1);
                    else if (message_part.plain_text)
                        message_part.body.append ("<br>");
                    else if (mime_type_is_text (message_part.mime_type))
                        message_part.body.append ("\n");
                    else
                        message_part.body.append (" ");
                }
            }
        }

        public void parse_text (GLib.Cancellable? cancellable=null) throws GLib.Error {
            if (parts == null)
                parse_body (cancellable);

            var child_parts = new List<MessagePart> ();
            foreach (var thread_part in parts) {
                /* Ignore empty parts inserted by faulty clients */
                if (thread_part.body.str.strip () == "")
                    continue;
                /* Select part, merge "inline" parts, commonly signatures */
                if (thread_part.mime_type == "text/html") {
                    if (html_part == null)
                        html_part = thread_part;
                    else if (thread_part.content_disposition == "inline") {
                        html_part.body.append (thread_part.body.str);
                        child_parts.remove (thread_part);
                    }
                }
                else if (thread_part.mime_type == "text/plain") {
                    if (html_part != null && thread_part.content_disposition == "inline") {
                        html_part.body.append (thread_part.body.str);
                        child_parts.remove (thread_part);
                    }
                    if (text_part == null)
                        text_part = thread_part;
                    else if (thread_part.content_disposition == "inline") {
                        text_part.body.append (thread_part.body.str);
                        child_parts.remove (thread_part);
                    }
                }
            }
        }
    }
}

