/*
 Copyright (C) 2011 Christian Dywan <christian@twotoasts.de>

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 See the file COPYING for the full license text.
*/

public class Postler.Messages : Gtk.TreeView {
    Accounts accounts;
    Gtk.TreeStore store;
    Gtk.TreePath defer_select;
    Postler.Client? client = null;

    public Postler.Content content { get; set; }

    public FolderType selected_folder_type { get; private set; }
    public string? location { get; private set; }
    public Message? selected_message { get; private set; }

    enum Columns {
        MESSAGE
    }

    bool search_inline (Gtk.TreeModel model, int column, string key,
        Gtk.TreeIter iter) {
        Message message;
        model.get (iter, Columns.MESSAGE, out message);
        return !(key in message.subject.down ());
    }

    void selection_changed () {
        Gtk.TreeIter iter;
        if (get_selected_iter (out iter)) {
            Message message;
            model.get (iter, Columns.MESSAGE, out message);
            selected_message = message;
        } else
            selected_message = null;
    }

    internal static string fallback_to_utf8 (string text) {
        var fallback = new StringBuilder.sized (text.size () * 2);
        for (unowned string s = text; s.get_char () != 0 ; s = s.next_char ()) {
            unichar character = s.get_char ();
            if (character.iscntrl () || !character.validate ())
                fallback.append_unichar ('�');
            else
                fallback.append_unichar (character);
        }
        return fallback.str;
    }

    void render_flag (Gtk.CellLayout layout, Gtk.CellRenderer cell,
        Gtk.TreeModel model, Gtk.TreeIter iter) {
        Message message;
        var screen = get_screen ();
        unowned string stock_id = null;
        unowned string prelight_stock_id = null;

        model.get (iter, Columns.MESSAGE, out message);
        if (message.flagged) {
            if (Gtk.IconTheme.get_for_screen (screen).has_icon (STOCK_STARRED))
                stock_id = STOCK_STARRED;
            else
                stock_id = STOCK_EMBLEM_IMPORTANT;
        } else {
            if (Gtk.IconTheme.get_for_screen (screen).has_icon (STOCK_NOT_STARRED))
                prelight_stock_id = STOCK_NOT_STARRED;
            if (message.forwarded)
                stock_id = STOCK_MAIL_FORWARDED;
            else if (message.replied)
                stock_id = STOCK_MAIL_REPLIED;
        }
        cell.set ("active", message.flagged,
                  "stock-id", stock_id, "prelight-stock-id", prelight_stock_id);
    }

    void render_subject (Gtk.CellLayout layout, Gtk.CellRenderer cell,
        Gtk.TreeModel model, Gtk.TreeIter iter) {

        Message message;
        var renderer = cell as Postler.CellRendererConvo;

        model.get (iter, Columns.MESSAGE, out message);

        if (message.uri == "error:") {
            renderer.subject = message.subject;
            renderer.sender = renderer.date = renderer.text = "";
            renderer.replies = 0;
            renderer.priority = false;
            return;
        }

        if (message.uri == null) {
            new Client ().get_message.begin (message, MessageFlags.NONE,
                new Gtk.TreeRowReference (model, model.get_path (iter)));
            renderer.subject = "...";
            return;
        }

        renderer.subject = message.subject ?? _("No subject");
        renderer.date = Postler.Content.format_date (message.date);
        renderer.sender = Postler.Contact.name_from_string (message.sender);
        renderer.replies = (int)message.replies;
        renderer.priority = message.priority;
        renderer.text = message.excerpt + "…";
    }

    void renderer_flag_toggled (Gtk.CellRendererToggle renderer,
                                string                 path) {
        Gtk.TreeIter iter = Gtk.TreeIter ();
        if (!model.get_iter_from_string (out iter, path))
            return;

        toggle_message_flag (iter, MessageFlags.TOGGLE_FLAGGED, () => { });
    }

    void on_drag_data_get (Gdk.DragContext context,
        Gtk.SelectionData selection_data, uint info, uint time_) {

        Gtk.TreeIter iter;
        Gtk.TreeModel model;
        var rows = get_selection ().get_selected_rows (out model);
        string[] uris = null;
        foreach (Gtk.TreePath path in rows) {
            model.get_iter_from_string (out iter, path.to_string ());
            Message message;
            model.get (iter, Columns.MESSAGE, out message);
            uris += message.uri;
        }
        if (uris != null)
            selection_data.set_uris (uris);
    }

    public Messages (Accounts accounts) {
        /* Drag and Drop */
        Gtk.drag_source_set (this, Gdk.ModifierType.BUTTON1_MASK, {}, Gdk.DragAction.MOVE);
        Gtk.drag_source_add_uri_targets (this);
        drag_data_get.connect (on_drag_data_get);

        this.accounts = accounts;
        set_fixed_height_mode (true);
        store = new Gtk.TreeStore (1, typeof (Message));
        set_search_equal_func (search_inline);
        get_selection ().set_mode (Gtk.SelectionMode.MULTIPLE);
        get_selection ().changed.connect (selection_changed);

        var column = new Gtk.TreeViewColumn ();
        column.set_sizing (Gtk.TreeViewColumnSizing.FIXED);
        column.set_title (_("Flagged"));
        var renderer_flag = new Postler.CellRendererToggle ();
        column.pack_start (renderer_flag, true);
        column.set_cell_data_func (renderer_flag, render_flag);
        renderer_flag.toggled.connect (renderer_flag_toggled);
        insert_column (column, -1);
        int cell_width, xpad;
        column.cell_get_size (null, null, null, out cell_width, null);
        Gtk.cell_renderer_get_padding (renderer_flag, out xpad, null);
        column.set_fixed_width (cell_width + xpad * 4);

        var renderer_text = new Postler.CellRendererConvo ();
        column = new Gtk.TreeViewColumn.with_attributes ("Messages", renderer_text, null);;
        column.set_sizing (Gtk.TreeViewColumnSizing.FIXED);
        column.set_title (_("Subject"));
        column.set_cell_data_func (renderer_text, render_subject);
        insert_column (column, -1);

        unowned Gtk.BindingSet binding_set = Gtk.BindingSet.by_class (get_class ());
        Gtk.BindingEntry.add_signal (binding_set,
            Gdk.keyval_from_name ("Return"), 0,
                "content-display", 0);
        Gtk.BindingEntry.add_signal (binding_set,
            Gdk.keyval_from_name ("ISO_Enter"), 0,
                "content-display", 0);
        Gtk.BindingEntry.add_signal (binding_set,
            Gdk.keyval_from_name ("KP_Enter"), 0,
                "content-display", 0);
        Gtk.BindingEntry.add_signal (binding_set,
            Gdk.keyval_from_name ("Return"), Gdk.ModifierType.CONTROL_MASK,
                "content-display-window", 0);
        Gtk.BindingEntry.add_signal (binding_set,
            Gdk.keyval_from_name ("ISO_Enter"), Gdk.ModifierType.CONTROL_MASK,
                "content-display-window", 0);
        Gtk.BindingEntry.add_signal (binding_set,
            Gdk.keyval_from_name ("KP_Enter"), Gdk.ModifierType.CONTROL_MASK,
                "content-display-window", 0);
        Gtk.BindingEntry.add_signal (binding_set,
            Gdk.keyval_from_name ("space"), 0,
                "content-page-down", 0);
        Gtk.BindingEntry.add_signal (binding_set,
            Gdk.keyval_from_name ("space"), Gdk.ModifierType.SHIFT_MASK,
                "content-page-up", 0);
        Gtk.BindingEntry.add_signal (binding_set,
            Gdk.keyval_from_name ("Delete"), 0, "delete", 0);
        content_display_window.connect (content_new_window);
    }

    [Signal (action=true)]
    public virtual signal void content_display () {
        select_cursor_row (false);
    }
    [Signal (action=true)]
    public signal void content_display_window ();
    /* Vala fails to parse out parameters in virtual signals */
    void content_new_window () {
        Gtk.TreeIter iter;
        if (get_selected_iter (out iter)) {
            toggle_message_flag (iter, MessageFlags.MARK_READ, (message, result) => {
                Postler.App.spawn_module ("content", (message as Message).get_path ());
            });
        }
    }
    [Signal (action=true)]
    public virtual signal void content_page_down () {
        content.move_cursor (Gtk.MovementStep.PAGES, 1);
    }
    [Signal (action=true)]
    public virtual signal void content_page_up () {
        content.move_cursor (Gtk.MovementStep.PAGES, -1);
    }
    [Signal (action=true)]
    public virtual signal void delete () {
        delete_selected ();
    }

    internal static string parse_encoded (string? encoded, out string charset) {
        charset = null;
        return_val_if_fail (encoded != null, null);

        /* format "=?charset?encoding?encoded?=",
           if in doubt, bail out and take the raw data */
        var decoded = new GLib.StringBuilder ();
        Regex regex;
        try {
            regex = new Regex("""((?<=\?=)|^)(?P<UNENCODED>.*?)(=\?(?P<CHARSET>[^?]*)\?(?P<ENCODING>[qQbB])\?(?P<ENCODED>.*?)\?=|$)""",
                                  GLib.RegexCompileFlags.DOTALL,0);
        }
        catch (GLib.RegexError e) {
            GLib.critical ("parse_encoded: could not parse regular expression");
            return encoded;
        }
         
        MatchInfo mi;
        if ( !regex.match (encoded, 0, out mi)) {
            return encoded;
        }

        try {
            do {
                var unencoded = mi.fetch_named ("UNENCODED") ?? "";
                charset = (mi.fetch_named ("CHARSET") ?? "").up ();
                var encoding = (mi.fetch_named ("ENCODING") ?? "").up ();
                var encoded_text = mi.fetch_named ("ENCODED") ?? "";

                if(charset == "" || encoding == "") {
                    decoded.append (unencoded);
                    continue;
                }      
                
                /* Encoding aliases */
                if (charset == "KS_C_5601-1987")
                    charset = "CP949";

                string unquoted;
                if (encoding == "Q") {
                    unquoted = encoded_text.replace (" =", " ").replace ("_", " ");
                    unquoted = Postler.Message.quoted_printable_decode (unquoted);
                }
                else if (encoding == "B") 
                    unquoted = (string)GLib.Base64.decode (encoded_text);
                else
                    unquoted = encoded_text;
                try {
                    decoded.append (unencoded + GLib.convert (unquoted, -1, "UTF-8", charset, null));
                }
                catch (GLib.ConvertError error) {
                    GLib.warning (_("Failed to convert \"%s\": %s"), encoded, error.message);
                    decoded.append (unencoded + encoded_text);
                }
            } while (mi.next ());
        }
        catch (GLib.RegexError e) {
            GLib.warning ("parse: could not parse encoded string starting with %s", encoded[0:40]);
            return encoded;
        }
    
        return fallback_to_utf8 (decoded.str);
    }

    public void clear () {
        client = null;
        store.clear ();
    }

    static string flags_from_filename (string filename, out string bare_filename) {
        string[]? parts = filename.split (":");
        if (parts == null || parts[0] == null || parts[1] == null) {
            bare_filename = filename;
            return "";
        }
        bare_filename = parts[0];
        return parts[1];
    }

    public async bool populate (string location, FolderType folder_type=FolderType.GENERIC) {

        if (this.location == location)
            return false;

        clear ();
        selected_folder_type = folder_type;
        var messages = new GLib.List<Message> ();

        try {
            client = new Client ();
            /* FIXME: The client emits, the callback isn't called */
            client.got_message.connect (got_message);
            messages = yield client.get_messages (location);

            foreach (var message in messages) {
                store.insert_with_values (null, null, 0,
                    Columns.MESSAGE, message);
            }

            /* Show error for failed search, ie. no results */
            int n_messages = store.iter_n_children (null);
            if (location.has_prefix ("search:") && n_messages == 0) {
                display_error (_("No messages found"),
                    _("Check the spelling or try a different filter."));
                get_selection().set_select_function( () => { return false; });
                return false;
            } else {
                get_selection().set_select_function( () => { return true; });
            }

            model = store;
            hadjustment.value = vadjustment.value = 0;

        } catch (GLib.Error error) {
            display_error (_("Failed to read folder \"%s\".").printf (location),
                error.message);
        }

        this.location = location;
        return false;
    }

    public void display_error (string title, string text) {
        clear ();
        string markup = "%s\n%s".printf (title, text);
        var message = new Message.from_id ("error:");
        message.from_error (markup);
        store.insert_with_values (null, null, 0,
                                  Columns.MESSAGE, message);
    }

    internal static string toggle_flag (string location, char flag) {
        string folder = Path.get_dirname (location);
        string bare_filename;
        if (folder.has_suffix ("/new"))
            folder = folder.slice (0, -4) + "/cur";
        string flags = flags_from_filename (location, out bare_filename);
        var new_flags = new StringBuilder ();
        bool did_include_flag = false;
        foreach (var character in flags.to_utf8 ()) {
            if (character < flag)
                new_flags.append_c (character);
            else if (character == flag)
                did_include_flag = true;
            else if (character > flag) {
                if (!did_include_flag) {
                    did_include_flag = true;
                    new_flags.append_c (flag);
                }
                new_flags.append_c (character);
            }
        }
        if (!did_include_flag)
            new_flags.append_c (flag);
        return folder + "/" + Path.get_basename (bare_filename) + ":" + new_flags.str;
    }

    bool get_selected_iter (out Gtk.TreeIter iter) {
        GLib.List<Gtk.TreePath> paths = get_selection ().get_selected_rows (null);
        iter = Gtk.TreeIter ();
        var path = paths.nth_data (0);
        if (path != null && model.get_iter (out iter, path))
            return true;
        return false;
    }

    bool get_message_iter (string location, out Gtk.TreeIter iter) {
        Gtk.TreeIter message_iter;
        if (!store.iter_children (out message_iter, null))
            return false;
        do {
            Message existing_message;
            store.get (message_iter, Columns.MESSAGE, out existing_message);
            if (existing_message.get_path () == location) {
                if (&iter != null)
                    iter = message_iter;
                return true;
            }
        } while (store.iter_next (ref message_iter));
        return false;
    }

    bool iter_previous (ref Gtk.TreeIter iter) {
        Gtk.TreePath path = model.get_path (iter);
        return path.prev () && model.get_iter (out iter, path);
    }

    public void select_next_unread (bool forward) {
        Gtk.TreeIter iter;
        var path = get_selection ().get_selected_rows (null).nth_data (0);
        if (path != null)
            model.get_iter (out iter, path);
        else if (forward)
            model.iter_children (out iter, null);
        else
            model.iter_nth_child (out iter, null, model.iter_n_children (null) - 1);
        while ((forward && model.iter_next (ref iter))
           || (!forward && iter_previous (ref iter))) {
            Message message;
            model.get (iter, Columns.MESSAGE, out message);
            if (message.replies > 0) {
                set_cursor (model.get_path (iter), null, false);
                break;
            }
        }
    }

    void got_message (Message message, int64 position) {
        if (get_message_iter (message.get_path (), null))
            return;
        /* TODO: Honor folder and filter */
        bool scroll = vadjustment.value == 0;
        store.insert_with_values (null, null, 0,
            Columns.MESSAGE, message);
        if (scroll)
            vadjustment.value = 0;
        /* if (get_message_iter (file.get_path (), out iter))
                store.remove (iter); */
    }

    public void toggle_selected_flag (char flag) {
        GLib.List<Gtk.TreePath> paths;
        var references = new GLib.List<Gtk.TreeRowReference> ();

        paths = get_selection ().get_selected_rows (null);
        foreach (var path in paths)
            references.prepend (new Gtk.TreeRowReference (model, path));

        foreach (var reference in references) {
            var path = reference.get_path ();
            Gtk.TreeIter iter;
            if (model.get_iter (out iter, path)) {
                toggle_message_flag (iter, flag, () => { });
            }
        }
    }

    void display_message (Gtk.TreeIter iter) {
        toggle_message_flag (iter, MessageFlags.MARK_READ, (message, result) => {
            content.display (message as Message);
        });
    }

    void toggle_message_flag (Gtk.TreeIter iter, char flag, AsyncReadyCallback callback) {
        Message message;
        model.get (iter, Columns.MESSAGE, out message);
        new Client ().get_message.begin (message, flag,
            new Gtk.TreeRowReference (model, model.get_path (iter)),
            (client, result) => { callback (message, result); });
    }

    public static bool ensure_folder (string folder) {
        if (DirUtils.create_with_parents (folder + "/new", 0700) != 0
         || DirUtils.create_with_parents (folder + "/cur", 0700) != 0)
            return false;
        return true;
    }

    public void delete_selected () {
        if (selected_folder_type == FolderType.TRASH)
            move_selected (null);
        else
            move_selected (FolderType.TRASH);
    }

    public static uint32 obtain_uid_sequence (string folder, uint32 fallback_sequence) {
        /* Update .uidvalidity number, also read and updated by mbsync */
        uint32 uid_sequence = fallback_sequence;
        string uid_file = folder.slice (0, -4) + ".uidvalidity";
        try {
            string contents;
            FileUtils.get_contents (uid_file, out contents, null);
            string[] parts = contents.split ("\n");
            if (parts[0] != null && parts[1] != null) {
                uint32 old_uid_sequence = (uint32)(parts[1].to_uint64 ());
                FileUtils.set_contents (uid_file, "%s\n%u\n".printf (
                    parts[0], old_uid_sequence + 1), -1);
                uid_sequence = old_uid_sequence + 1;
            }
        }
        catch (Error error) {
            GLib.critical (_("Failed to read folder validity: %s"), error.message);
        }
        return uid_sequence;
    }

    static uint32 maildir_sequence_number = 0;
    public static string generate_filename (string folder, string flags="")
        requires (folder.has_suffix ("/cur/")) {
        /* format "time.pid_sequence.host,U=uid:2,FLAGS"
           This is the format used by mbsync. */
        var now = GLib.TimeVal ();
        now.get_current_time ();
        if (maildir_sequence_number == uint32.MAX)
            maildir_sequence_number = 0;
        else
            maildir_sequence_number++;
        uint32 uid_sequence = obtain_uid_sequence (folder, maildir_sequence_number);
        return folder + "%lu.%d_%u.%s,U=%u:2,%s".printf (
            now.tv_sec,
            Posix.getpid (),
            maildir_sequence_number,
            Environment.get_host_name ().replace ("/", "_").replace (":", "."),
            uid_sequence,
            flags);
    }

    public static string update_filename (string location, string new_folder)
        requires (new_folder.has_suffix ("/cur/")) {

        string bare_filename;
        string flags = flags_from_filename (location, out bare_filename);
        string filename;
        if ("S" in flags)
            filename = Path.get_basename (bare_filename) + ":" + flags;
        else
            filename = Path.get_basename (toggle_flag (location, MessageFlags.TOGGLE_READ));
        string[] parts = filename.split (",");
        return_val_if_fail (parts[0] != null && parts[1] != null
            && parts[2] != null && parts[3] == null, "");
        uint32 uid_sequence = obtain_uid_sequence (new_folder, maildir_sequence_number);
        return new_folder + parts[0] + ",U=" + uid_sequence.to_string () + ":2," + parts[2];
    }

    public void move_selected (FolderType? type) {
        string? destination_path;
        if (type == null)
            destination_path = null;
        else
            destination_path = "/"; /* Nonsense path, in case of later errors */

        GLib.List<Gtk.TreePath> paths;
        var references = new GLib.List<Gtk.TreeRowReference> ();
        paths = get_selection ().get_selected_rows (null);
        foreach (var path in paths)
            references.prepend (new Gtk.TreeRowReference (model, path));

        foreach (var reference in references) {
            var path = reference.get_path ();
            Gtk.TreeIter iter;
            if (model.get_iter (out iter, path)) {
                Message message;
                model.get (iter, Columns.MESSAGE, out message);
                try {
                    AccountInfo account_info = message.get_account (accounts);
                    return_if_fail (account_info != null);
                    string? folder_name = null;
                    foreach (var info in accounts.get_infos ()) {
                        if (info.path == account_info.path) {
                            folder_name = info.get_folder (type);
                            break;
                        }
                    }
                    destination_path = account_info.path + "/" + folder_name;
                    return_if_fail (destination_path != null);
                    message.move (destination_path);

                    /* Move to next, more recent message row */
                    Gtk.TreePath next_path = model.get_path (iter);
                    content.clear ();
                    if (store.remove (iter)) {
                        next_path.prev ();
                        if (model.get_iter (out iter, next_path)) {
                            set_cursor (model.get_path (iter), null, false);
                            display_message (iter);
                        }
                    }
                } catch (GLib.Error error) {
                    unowned string warning;
                    if (destination_path == null)
                        warning = _("Failed to delete message \"%s\": %s");
                    else
                        warning = _("Failed to move message \"%s\": %s");
                    GLib.critical (warning, location, error.message);
                }
            }
        }
    }

    public override bool button_press_event (Gdk.EventButton event) {
        Gtk.TreePath path;
        Gtk.TreeViewDropPosition pos;
        get_dest_row_at_pos ((int)event.x, (int)event.y, out path, out pos);

        if (event.type == Gdk.EventType.2BUTTON_PRESS) {
            Gtk.TreeIter iter;
            if (get_selected_iter (out iter)) {
                toggle_message_flag (iter, MessageFlags.MARK_READ,
                    (message, result) => {
                    Postler.App.spawn_module ("content", (message as Message).get_path ());
                });
            }
        } else if (event.type == Gdk.EventType.BUTTON_PRESS
                && get_selection ().count_selected_rows () > 1
                && ((event.state & (Gdk.ModifierType.CONTROL_MASK
                                  | Gdk.ModifierType.SHIFT_MASK)) == 0)
                && get_selection ().path_is_selected (path)) {
                    get_selection().set_select_function( () => { return false; });
                    defer_select = path;
        } else {
            /* Re-enable selection */
            get_selection().set_select_function(() => { return true; });
            if (defer_select != null
             && get_dest_row_at_pos ((int)event.x, (int)event.y, out path, out pos)
             && defer_select == path
             && !(event.x == 0 && event.y == 0))
                set_cursor (path, null, false);

            defer_select = null;
        }
        return base.button_press_event (event);
    }

    public override bool button_release_event (Gdk.EventButton event) {
        if ((event.state & Gdk.ModifierType.BUTTON1_MASK) == 0) {
            /* Re-enable selection */
            get_selection().set_select_function(() => { return true; });
            Gtk.TreePath? path = null;
            Gtk.TreeViewDropPosition pos;
            if (defer_select != null
             && get_dest_row_at_pos ((int)event.x, (int)event.y, out path, out pos)
             && defer_select == path
             && !(event.x == 0 && event.y == 0))
                set_cursor (path, null, false);

            defer_select = null;
            return base.button_release_event (event);
        } else {
            get_selection ().set_select_function(() => { return true; });
            Gtk.TreePath? path = null;
            Gtk.TreeViewDropPosition pos;
            get_dest_row_at_pos ((int)event.x, (int)event.y, out path, out pos);
            if (defer_select != null
             && get_dest_row_at_pos ((int)event.x, (int)event.y, out path, out pos)
             && !(event.x == 0 && event.y == 0))
                set_cursor (path, null, false);

            defer_select = null;
        }

        GLib.List<Gtk.TreePath> paths = get_selection ().get_selected_rows (null);
        Gtk.TreeIter iter = Gtk.TreeIter ();
        var path = paths.nth_data (0);
        if (path != null && model.get_iter (out iter, path)) {
            Gtk.TreeViewColumn column;
            if (get_path_at_pos ((int)event.x, (int)event.y, null,
                                  out column, null, null)) {
                unowned string title = column.get_title ();
                /* Clickable icons should not display/ mark as read */
                if (title != _("Status") && title != _("Flagged"))
                    display_message (iter);
            }
        }
        return base.button_release_event (event);
    }

    public override void row_activated (Gtk.TreePath path, Gtk.TreeViewColumn column) {
        Gtk.TreeIter iter;
        if (model.get_iter (out iter, path))
            display_message (iter);
    }
}

