/*
 Copyright (C) 2010 Christian Dywan <christian@twotoasts.de>

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 See the file COPYING for the full license text.
*/

public class Postler.Reader {
    static string module = null;
    static bool verbose = false;
    static bool version = false;
    static string[] filenames = null;
    const OptionEntry[] options = {
        { "module", 'm', 0, OptionArg.STRING, ref module, N_("Module"), null },
        { "verbose", 'v', 0, OptionArg.NONE, ref verbose, N_("Verbose"), null },
        { "version", 'V', 0, OptionArg.NONE, ref version, N_("Display program version"), null },
        { "", 0, 0, GLib.OptionArg.STRING_ARRAY, ref filenames, N_("Filenames"), null },
        { null }
    };

    static int main (string[] args) {
        Postler.App.argv0 = args[0];

#if HAVE_NLS
        Intl.bindtextdomain (Config.GETTEXT_PACKAGE, Config.LOCALEDIR);
        Intl.bind_textdomain_codeset (Config.GETTEXT_PACKAGE, "UTF-8");
        Intl.textdomain (Config.GETTEXT_PACKAGE);
#endif

        var context = new GLib.OptionContext (
            /* i18n: The meaning of the uppercase values is analogous to the
                     fields at the top of a message, ie. From, To, Copy */
            _("[mailto:][ADDRESS][?subject=SUBJECT][&body=BODY]")
          + _("[&from=FROM][&to=TO][&cc=COPY][&attach=ATTACHMENT][&part=PART]"));
        context.set_help_enabled (true);
        context.add_main_entries (options, null);
        context.add_group (Gtk.get_option_group (false));
        Environment.set_application_name (_("Postler"));
        try {
	    context.parse (ref args);
	} catch (GLib.Error error) {
            /* i18n: Command line arguments are invalid */
	    GLib.error (_("Failed to parse command line: %s"), error.message);
	}

        if (version) {
            stdout.printf ("%s %s\n\n%s\n\t%s\n",
                Config.PACKAGE_NAME, Config.PACKAGE_VERSION,
                _("Please report comments, suggestions and bugs to:"),
                Config.PACKAGE_BUGREPORT);
            return 0;
        }

        /* Colored log domains according to level and less visual noise */
        LogFunc handler = (domain, level, message) => {
            int color;
            switch (level) {
            case LogLevelFlags.LEVEL_DEBUG:
                color = 2; /* green */
                break;
            case LogLevelFlags.LEVEL_MESSAGE:
            case LogLevelFlags.LEVEL_INFO:
                color = 4; /* blue */
                break;
            case LogLevelFlags.LEVEL_WARNING:
                color = 3; /* yellow */
                break;
            case LogLevelFlags.LEVEL_CRITICAL:
            case LogLevelFlags.LEVEL_ERROR:
                color = 1; /* red */
                break;
            default:
                color = 7; /* white */
                break;
            }
            stdout.printf ("\x001b[%dm [%s]\x001b[0m %s\n",
                color + 30 + 60,
                domain ?? Environment.get_prgname ().to_string (),
                message);
        };
        string[] domains = { null, "GLib", "GLib-GObject", "GLib-GIO",
                             "Gdk", "Gdk-Pixbuf", "Pango", "Gtk" };
        foreach (var domain in domains)
            Log.set_handler (domain, LogLevelFlags.LEVEL_MASK
                                   | LogLevelFlags.FLAG_FATAL
                                   | LogLevelFlags.FLAG_RECURSION, handler);

        /* --verbose or POSTLER_DEBUG=1 enables debug messages */
        unowned string? debug = Environment.get_variable ("POSTLER_DEBUG");
        if (verbose || debug == "1") {
            Environment.set_variable ("POSTLER_DEBUG", "1", false);
        }
        else {
            handler = (domain, level, message) => { };
            Log.set_handler (null, LogLevelFlags.LEVEL_DEBUG, handler);
        }

        /* mailto, ? or @ implies compose, otherwise file implies content */
        if (module == null && filenames != null && filenames[0] != null) {
            if (filenames[0].has_prefix ("mailto:")
             || filenames[0].has_prefix ("?")
             || (filenames[0].chr (-1, '@') != null && filenames[0].chr (-1, '/') == null))
                module = "compose";
            else
                module = "content";
        }

        if (module == null)
            module = "bureau";

        Environment.set_prgname ("postler-" + module);

        if (module == "service") {
            var service = new Postler.Service ();
            service.done.connect ((status) => {
                Process.exit (status);
            });
            service.run ();
            assert_not_reached ();
        }

        Gtk.init (ref args);
        /* FIXME: only for new webkit */
        WebKit.set_cache_model (WebKit.CacheModel.DOCUMENT_VIEWER);
        Postler.App.register_stock_items ();

        if (module != "bureau") {
            if (module == "content" || module == "source") {
                var window = new Gtk.Window (Gtk.WindowType.TOPLEVEL);
                window.set_icon_name (module == "source" ? "text-html" : "emblem-mail");
                var screen = window.get_screen ();
                Gdk.Rectangle monitor;
                screen.get_monitor_geometry (0, out monitor);
                double window_width = monitor.width / 2.2;
                window.set_default_size ((int)window_width, (int)(monitor.height / 2.2));
                window.show_all ();
                /* Show toplevel early so stylesheets can be themed */
                var shelf = new Gtk.VBox (false, 0);
                window.add (shelf);
                var content = new Postler.Content ();
                shelf.add (new Postler.Viewer (content));

                if (filenames != null && filenames[0] != null) {
                    Message message;
                    try {
                        string filename = filenames[0];
                        if (filename.has_prefix ("file://"))
                            message = new Message.from_file (GLib.File.new_for_uri (filename));
                        else if (FileUtils.test (filename, FileTest.EXISTS))
                            message = new Message.from_file (GLib.File.new_for_path (filename));
                        else
                            message = new Message.from_id (filename);
                    } catch (GLib.Error error) {
                        GLib.error (error.message);
                    }
                    content.notify["subject"].connect ((object, pspec) => {
                        window.set_title (content.subject);
                    });
                    if (module == "source")
                        content.display_source (message);
                    else
                        content.display (message);
                }

                shelf.show_all ();
                window.destroy.connect (Gtk.main_quit);
                Gtk.main ();
                return 0;
            }
            else if (module == "compose") {
                Postler.Composer composer = new Postler.Composer ();
                /* Show toplevel early so stylesheets can be themed */
                composer.show ();

                string mailto = "";
                if (filenames != null && filenames[0] != null)
                    mailto = filenames[0];
                /* mailto: may or may not be stripped by the calling app */
                if (mailto.has_prefix ("mailto:"))
                    mailto = mailto.substring (7, -1);
                /* A single leading ? is valid and should be ignored */
                if (mailto.has_prefix ("?"))
                    mailto = mailto.substring (1, -1);

                    string[] fields = mailto.split_set ("?&");
                    foreach (var field in fields) {
                        string[] pieces = field.split ("=");
                        bool success;
                        if (pieces[0] != null && pieces[1] != null)
                            success = composer.add_field (pieces[0], Soup.URI.decode (pieces[1]));
                        else
                            success = composer.add_field ("to", Soup.URI.decode (field));
                        if (!success)
                            GLib.warning (_("Invalid field \"%s\" was ignored."),
                                          pieces[1] != null ? pieces[0] : field);
                    }

                    if (filenames != null && filenames[1] != null) {
                        composer.prepare_reply (filenames[1],
                            composer.subject.has_prefix ("Re: ")
                         || composer.subject.has_prefix ("Fw: "));
                    } else {
                        composer.prepare_reply ();
                    }

                composer.destroy.connect (Gtk.main_quit);
                Gtk.main ();
                return 0;
            }
            else
                GLib.error ("Unknown module \"%s\". Valid modules are: %s",
                    module, "bureau content source compose service");
        }

        var app = new Postler.App ();
#if HAVE_GTK3
        try {
            app.register ();
            if (app.is_remote) {
                app.activate ();
                return 0;
            }
        }
        catch (GLib.Error error) {
            GLib.error (_("Failed to activate running instance"));
        }
#else
        if (app.is_running) {
            Unique.MessageData data = new Unique.MessageData ();
            if (filenames != null && filenames[0] != null)
                data.set_text (filenames[0], filenames[0].length);
            var response = app.send_message (Unique.Command.ACTIVATE, data);
            if (response == Unique.Response.OK)
                return 0;
            GLib.error (_("Failed to activate running instance"));
        }
#endif

        var bureau = new Postler.Bureau ();
        bureau.destroy.connect (Gtk.main_quit);
        bureau.show ();
        if (filenames != null && filenames[0] != null) {
            GLib.Idle.add (() => {
                bureau.folders.select_folder (filenames[0]);
                return false;
            });
        }
        app.watch_window (bureau);
        Gtk.main ();
        return 0;
    }
}

